###############################################################################
# Name: html.py                                                               #
# Purpose: Define HTML syntax for highlighting and other features             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: html.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for HTML/DHTML/SGML.
@todo: Add Netscape/Microsoft Tag Extenstions (maybe)
@todo: Styleing needs lots of tweaking

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _html.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata
import _javascript
import _vbscript

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

# HTML Tags (HTML4)
HTML_TAGS = (0, "address applet area a base basefont big blockquote br caption "
                "center cite code dd dfn dir div dl dt font form hr html img "
                "input isindex kbd li link map menu meta ol option param pre p "
                "samp span select small strike sub sup table td textarea th tr "
                "script noscript tt ul var xmp b i u h1 h2 h3 h4 h5 h6 em "
                "strong head body title "
                # HTML 4.0 Tags
                "abbr acronym bdo button col label colgroup del fieldset "
                "iframe ins legend object optgroup q s tbody tfoot thead "
                # HTML 5 Tags
                "article aside audio canvas command datalist details dialog "
                "embed figcaption figure footer header hgroup keygen mark "
                "meter nav output progress rp rt ruby section source time "
                "video "
                # Tag Attributes / Arguments
                "action align alink alt archive background bgcolor border "
                "bordercolor cellpadding cellspacing checked class clear "
                "codebase color cols colspan content coords enctype face "
                "gutter height hspace id link lowsrc marginheight marginwidth "
                "maxlength method name prompt rel rev rows rowspan scrolling "
                "selected shape size src start target text type url usemap "
                "ismap valign value vlink vspace width wrap href http-equiv "
                # HTML 4 Tag Attributes /Arguments
                "accept accesskey axis char charoff charset cite classid "
                "codetype compact data datetime declare defer dir disabled for "
                "frame headers hreflang lang language longdesc multiple nohref "
                "nowrap profile readonly rules scheme scope standby style "
                "summary tabindex valuetype version "
                # HTML 5 Tag Attributes / Arguments
                "async autocomplete contenteditable contextmenu date "
                "datetime-local draggable email formaction formenctype "
                "formmethod formnovalidate formtarget hidden list manifest max "
                "media min month novalidate number pattern ping range required "
                "reversed role sandbox scoped seamless search sizes spellcheck "
                "srcdoc step tel week "
                # DHTML Support
                "dtml-var dtml-if dtml-unless dtml-in dtml-with dtml-let "
                "dtml-call dtml-raise dtml-try dtml-comment dtml-tree")

#---- Extra defs ----#
# ColdFusion Tags
CF_TAGS = ("cfabort cfapplet cfapplication cfargument cfassociate cfbreak "
           "cfcache cfcalendar cfcase cfcatch cfchart cfchartdata "
           "cfchartseries cfcol cfcollection cfcomponent cfcontent cfcookie "
           "cfdefaultcase cfdirectory cfdocument cfdocumentitem "
           "cfdocumentsection cfdump cfelse cfelseif cferror cfexecute cfexit "
           "cffile cfflush cfform cfformgroup cfformitem cfftp cffunction "
           "cfgrid cfgridcolumn cfgridrow cfgridupdate cfheader cfhtmlhead "
           "cfhttp cfhttpparam cfif cfimport cfinclude cfindex cfinput "
           "cfinsert cfinvoke cfinvokeargument cfldap cflocation cflock cflog "
           "cflogin cfloginuser cflogout cfloop cfmail cfmailparam cfmailpart "
           "cfmodule cfNTauthenticate cfobject cfobjectcache cfoutput cfparam "
           "cfpop cfprocessingdirective cfprocparam cfprocresult cfproperty "
           "cfquery cfqueryparam cfregistry cfreport cfreportparam cfrethrow "
           "cfreturn cfsavecontent cfschedule cfscript cfsearch cfselect cfset "
           "cfsetting cfsilent cfslider cfstoredproc cfswitch cftable "
           "cftextarea cfthrow cftimer cftrace cftransaction cftree cftreeitem "
           "cftry cfupdate cfwddx cfxml")

# JavaScript Keywords (see javascript.py)
JS_KEYWORDS = (1, _javascript.KeywordString(synglob.ID_LANG_JS))

# VBScript Keywords (currently unsupported)
VBS_KEYWORDS = (2, _vbscript.VBS_KW)

# Python Keywords (see python.py)
PY_KEYWORDS = (3, "")

# PHP Keywords (see php.py)
# This module is loaded for files with a .html/htm extension so it is assumed
# that there is no php in the file. On the other hand the php module loads
# this module so that it can support embedded html. This behavior may be changed
# in the future

# XML Keywords (see xml.py)
# XML files are handled independantly from html although there is support for
# embedded xml highlighting it is currently not being used.

# SGML Keywords
SGML_KEYWORDS = (5, "ELEMENT DOCTYPE ATTLIST ENTITY NOTATION")
#SGML_KEYWORDS = (5, "#CURRENT #IMPLIED #REQUIRED ATTLIST CDATA DOCTYPE "
#                    "ELEMENT ENTITY HTML IDREF INCLUDE IGNORE NMTOKEN NUMBER "
#                    "RCDATA TEMP")

# SGML Block Keywords
SGML_BLOCK = (7, "")

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_H_DEFAULT, 'default_style'),
                 (stc.STC_H_ASP, 'array_style'),
                 (stc.STC_H_ASPAT, 'array_style'),
                 (stc.STC_H_ATTRIBUTE, 'keyword2_style'),
                 (stc.STC_H_ATTRIBUTEUNKNOWN, 'error_style'),
                 (stc.STC_H_CDATA, 'default_style'), # Style ME
                 (stc.STC_H_COMMENT, 'comment_style'),
                 (stc.STC_H_DOUBLESTRING, 'string_style'),
                 (stc.STC_H_ENTITY, 'default_style'), # Style ME
                 (stc.STC_H_NUMBER, 'number_style'),
                 (stc.STC_H_OTHER, 'default_style'),  # Style ME
                 (stc.STC_H_QUESTION, 'scalar_style'),
                 (stc.STC_H_SCRIPT, 'funct_style'), # STYLE ME
                 (stc.STC_H_SGML_1ST_PARAM, 'keyword2_style'), # STYLE ME
                 (stc.STC_H_SGML_1ST_PARAM_COMMENT, 'comment_style'),
                 (stc.STC_H_SGML_BLOCK_DEFAULT, 'default_style'), # STYLE ME
                 (stc.STC_H_SGML_COMMAND, 'keyword_style'), # STYLE ME
                 (stc.STC_H_SGML_COMMENT, 'comment_style'),
                 (stc.STC_H_SGML_DEFAULT, 'array_style'), # STYLE ME
                 (stc.STC_H_SGML_DOUBLESTRING, 'string_style'),
                 (stc.STC_H_SGML_ENTITY, 'default_style'), # STYLE ME
                 (stc.STC_H_SGML_ERROR, 'error_style'),
                 (stc.STC_H_SGML_SIMPLESTRING, 'string_style'),
                 (stc.STC_H_SGML_SPECIAL, 'default_style'), # STYLE ME
                 (stc.STC_H_SINGLESTRING, 'string_style'),
                 (stc.STC_H_TAG, 'keyword_style'),
                 (stc.STC_H_TAGEND, 'keyword_style'),
                 (stc.STC_H_TAGUNKNOWN, 'error_style'),
                 (stc.STC_H_VALUE, 'number_style'),
                 (stc.STC_H_XCCOMMENT, 'comment_style'),
                 (stc.STC_H_XMLEND, 'scalar_style'),
                 (stc.STC_H_XMLSTART, 'scalar_style'),

                 # Embedded JavaScript
                 (stc.STC_HJ_COMMENT, 'comment_style'),
                 (stc.STC_HJ_COMMENTDOC, 'comment_style'),
                 (stc.STC_HJ_COMMENTLINE, 'comment_style'),
                 (stc.STC_HJ_DEFAULT, 'default_style'),
                 (stc.STC_HJ_DOUBLESTRING, 'default_style'), # STYLE ME
                 (stc.STC_HJ_KEYWORD, 'default_style'), # STYLE ME
                 (stc.STC_HJ_NUMBER, 'default_style'), # STYLE ME
                 (stc.STC_HJ_REGEX, 'default_style'), # STYLE ME
                 (stc.STC_HJ_SINGLESTRING, 'default_style'), # STYLE ME
                 (stc.STC_HJ_START, 'default_style'), # STYLE ME
                 (stc.STC_HJ_STRINGEOL, 'default_style'), # STYLE ME
                 (stc.STC_HJ_SYMBOLS, 'default_style'), # STYLE ME
                 (stc.STC_HJ_WORD, 'default_style'), # STYLE ME

                 (stc.STC_HJA_COMMENT, 'comment_style'),
                 (stc.STC_HJA_COMMENTDOC, 'comment_style'),
                 (stc.STC_HJA_COMMENTLINE, 'comment_style'),
                 (stc.STC_HJA_DEFAULT, 'default_style'),
                 (stc.STC_HJA_DOUBLESTRING, 'default_style'), # STYLE ME
                 (stc.STC_HJA_KEYWORD, 'default_style'), # STYLE ME
                 (stc.STC_HJA_NUMBER, 'default_style'), # STYLE ME
                 (stc.STC_HJA_REGEX, 'default_style'), # STYLE ME # STYLE ME
                 (stc.STC_HJA_SINGLESTRING, 'default_style'), # STYLE ME
                 (stc.STC_HJA_START, 'default_style'), # STYLE ME
                 (stc.STC_HJA_STRINGEOL, 'default_style'), # STYLE ME
                 (stc.STC_HJA_SYMBOLS, 'default_style'), # STYLE ME
                 (stc.STC_HJA_WORD, 'default_style'),
                 
                 (stc.STC_HBA_DEFAULT, 'operator_style'), # Styles ( ) ?
                 (stc.STC_HBA_COMMENTLINE, 'comment_style'),
                 (stc.STC_HBA_IDENTIFIER, 'default_style'), # TODO
                 (stc.STC_HBA_NUMBER, 'number_style'),
                 (stc.STC_HBA_START, 'default_style'), # TODO
                 (stc.STC_HBA_STRING, 'string_style'),
                 (stc.STC_HBA_STRINGEOL, 'stringeol_style'),
                 (stc.STC_HBA_WORD, 'keyword_style')  ]

#---- Extra Properties ----#
FOLD = ("fold", "1")
FLD_HTML = ("fold.html", "1")

#------------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Html and related languages""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_HTML)
        self.RegisterFeature(synglob.FEATURE_AUTOINDENT, AutoIndenter)

    def GetKeywords(self):
        """Returns Specified Keywords List"""
        if self.LangId == synglob.ID_LANG_COLDFUSION:
            return [(HTML_TAGS[0], HTML_TAGS[1] + " " + CF_TAGS), JS_KEYWORDS]
        else:
            return [HTML_TAGS, JS_KEYWORDS, SGML_KEYWORDS, VBS_KEYWORDS]

    def GetSyntaxSpec(self):
        """Syntax Specifications"""
        return SYNTAX_ITEMS + _javascript.SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set"""
        return [FOLD, FLD_HTML]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code"""
        return [u'<!--', u'-->']

#-----------------------------------------------------------------------------#

def AutoIndenter(estc, pos, ichar):
    """Auto indent python code.
    @param estc: EditraStyledTextCtrl
    @param pos: current carat position
    @param ichar: Indentation character

    """
    rtxt = u''
    line = estc.GetCurrentLine()
    spos = estc.PositionFromLine(line)
    text = estc.GetTextRange(spos, pos)
    eolch = estc.GetEOLChar()
    inspace = text.isspace()

    # Cursor is in the indent area somewhere
    if inspace:
        estc.AddText(eolch + text)
        return

    # Check if the cursor is in column 0 and just return newline.
    if not len(text):
        estc.AddText(eolch)
        return

    if ichar == u"\t":
        tabw = estc.GetTabWidth()
    else:
        tabw = estc.GetIndent()

    # Standard indent to match previous line
    indent = estc.GetLineIndentation(line)
    levels = indent / tabw
    end_spaces = ((indent - (tabw * levels)) * u" ")
    rtxt = eolch + (ichar * levels) + end_spaces

    # Check if we need some 'special' indentation
    tmp = text.rstrip()
    if tmp.endswith(u">"):
        # At a tag check for if we need extra indentation
        tagstart = tmp.rfind(u"<")
        if tagstart >= 0:
            tagval = tmp[tagstart:]
            if not tagval.startswith(u"</") and \
               not tagval.endswith(u"/>") and \
               not tagval.endswith(u"?>"):
                # Cursor is after an opening tag so we need to indent more
                # First match to the starting tag
                levels = (tagstart / tabw) # Add an extra level
                end_spaces = ((tagstart - (tabw * levels)) * u" ")
                rtxt = eolch + (ichar * (levels+1)) + end_spaces

    # Put text in the buffer
    estc.AddText(rtxt)

#---- Syntax Modules Internal Functions ----#
def KeywordString(option=0):
    """Returns the specified Keyword String
    @param option: specific subset of keywords to get

    """
    if option == synglob.ID_LANG_SGML:
        return SGML_KEYWORDS[1]
    else:
        return HTML_TAGS[1]

#---- End Syntax Modules Internal Functions ----#



###############################################################################
# Name: __init__.py                                                           #
# Purpose: Editra Business Model Library                                      #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Editra Business Model Library:

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__cvsid__ = "$Id: __init__.py 69798 2011-11-23 02:59:40Z CJP $"
__revision__ = "$Revision: 69798 $"

#-----------------------------------------------------------------------------#

# Text Utils
from searcheng import *
from fchecker import *
from fileutil import *
from _dirmon import *
from fileimpl import *
from txtutil import *
from logfile import *

from backupmgr import *
from calllock import *

# Storage Classes
from histcache import *
from clipboard import *

# Networking utilities
from e_weblib import *

# Misc
from miscutil import *
from _efactory import *
from cmenumgr import *
from efilehist import *
from osutil import *
from _threads import *
from _trash import *



###############################################################################
# Name: __init__.py                                                           #
# Purpose: Puts external modules in the namespace                             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################
"""External Module Package
Modules that Editra depends on that are where not developed for the project but
are distributed with it, either because of slight customizations made to them
or to reduce what is needed to be installed when installing Editra.

@note: modules in this directory are dependancies and addons that are not
       part of the core code.

"""
__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: __init__.py 52855 2008-03-27 14:53:06Z CJP $"
__revision__ = "$Revision: 52855 $"

__all__ = ['ez_setup', 'pkg_resources', 'events', 'flatnotebook']




###############################################################################
# Name: pike.py                                                               #
# Purpose: Define highlighting/syntax for Pike programming language           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: pike.py
@summary: Defines syntax and highlighting settings for the Pike programming
          language. Pike is very similar in form to C/CPP so the Cpp lexer is
          used to provide the highlighting settings.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _pike.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata
import _cpp

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
PIKE_KW = (0, "goto break return continue case default if else switch while "
              "foreach do gauge destruct lambda inherit import typeof catch "
              "for inline nomask")

PIKE_TYPE = (1, "private protected public static "
                "int string void float mapping array multiset mixed program "
                "object function")
#---- End Keyword Definitions ----#

#-----------------------------------------------------------------------------#

class SyntaxData(_cpp.SyntaxData):
    """SyntaxData object for Pike""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [PIKE_KW, PIKE_TYPE, _cpp.DOC_KEYWORDS]

    def GetCommentPattern(self):
        """Get the comment pattern"""
        return [u"//"]



###############################################################################
# Name: auinavi.py                                                            #
# Purpose: AuiMgr Pane navigator                                              #
# Author: Giuseppe "Cowo" Corbelli                                            #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Editra Control Library: AuiPaneNavigator

Popup navigation window for quickly navigating through AuiPanes in an AuiMgr.
Activating the dialog will cause a modal popup dialog with a list of all panes
managed by the aui manager. Changing the selection in the dialog will highlight
the pane in the managed frame. Selecting the choice in the list will move the
focus to that pane.

+---------------------+
| bmp title           |
+---------------------+
| pane list           |
|                     |                   
|                     |                   
|                     |                   
|                     |                   
|                     |
+---------------------+                   

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: auinavi.py 65794 2010-10-13 14:10:09Z CJP $"
__revision__ = "$Revision: 65794 $"

__all__ = ['AuiPaneNavigator',]

#-----------------------------------------------------------------------------#
# Imports
import wx

# Editra Control Libray Imports
import ctrlbox

#-----------------------------------------------------------------------------#

class AuiPaneNavigator(wx.Dialog):
    """Navigate through Aui Panes"""
    def __init__(self, parent, auiMgr, icon=None, title=''):
        """Initialize the navigator window
        @param parent: parent window
        @param auiMgr: wx.aui.AuiManager
        @keyword icon: wx.Bitmap or None
        @keyword title: string (dialog title)

        """
        super(AuiPaneNavigator, self).__init__(parent, wx.ID_ANY,
                                               "", style=wx.STAY_ON_TOP)

        # Attributes
        self._auimgr = auiMgr
        self._selectedItem = -1
        self._indexMap = list()
        self._sel = 0
        self._tabed = 0
        self._close_keys = [wx.WXK_ALT, wx.WXK_CONTROL, wx.WXK_RETURN]
        self._navi_keys = [wx.WXK_TAB, ord('1')] # <- TEMP
        self._listBox = None
        self._panel = None

        # Setup
        self.__DoLayout(icon, title)

        # Get the panes
        self.PopulateListControl()

        # Event Handlers
        self._listBox.Bind(wx.EVT_KEY_UP, self.OnKeyUp)
        self._listBox.Bind(wx.EVT_NAVIGATION_KEY, self.OnNavigationKey)
        self._listBox.Bind(wx.EVT_LISTBOX_DCLICK, self.OnItemSelected)
        self._listBox.Bind(wx.EVT_LISTBOX, lambda evt: self.HighlightPane())

    def __del__(self):
        self._auimgr.HideHint()

    def __DoLayout(self, icon, title):
        """Layout the dialog controls
        @param icon: wx.Bitmap or None
        @param title: string

        """
        sz = wx.BoxSizer(wx.VERTICAL)
        self._listBox = wx.ListBox(self, wx.ID_ANY, wx.DefaultPosition,
                                   wx.Size(200, 150), list(),
                                   wx.LB_SINGLE | wx.NO_BORDER)

        self._panel = ctrlbox.ControlBar(self,
                                         style=ctrlbox.CTRLBAR_STYLE_GRADIENT)
        self._panel.SetVMargin(2, 2)

        if icon is not None:
            bmp = wx.StaticBitmap(self._panel, bitmap=icon)
            self._panel.AddControl(bmp, wx.ALIGN_LEFT)

        txt = wx.StaticText(self._panel, label=title)
        self._panel.AddControl(txt, wx.ALIGN_LEFT)

        sz.Add(self._panel, 0, wx.EXPAND)
        sz.Add(self._listBox, 1, wx.EXPAND)
        sz.Fit(self)
        sz.SetSizeHints(self)
        sz.Layout()
        self.Centre()
        self.SetSizer(sz)
        self.SetAutoLayout(True)

    def OnKeyUp(self, event):
        """Handles wx.EVT_KEY_UP"""
        self._auimgr.HideHint()
        key_code = event.GetKeyCode()
        # TODO: add setter method for setting the navigation key
        if key_code in self._navi_keys:
            self._tabed += 1

            # Don't move selection on initial show
            if self._tabed == 1:
                self.HighlightPane()
                event.Skip()
                return

            selected = self._listBox.GetSelection() + 1
            if selected >= self._listBox.GetCount():
                selected = 0

            self._listBox.SetSelection(selected)
            self.HighlightPane()
            event.Skip()
        elif key_code in self._close_keys:
            self.CloseDialog()
        elif key_code == wx.WXK_ESCAPE:
            self.CloseDialog()
        else:
            event.Skip()

    def OnNavigationKey(self, event):
        """Handles wx.EVT_NAVIGATION_KEY"""
        selected = self._listBox.GetSelection()
        maxItems = self._listBox.GetCount()
            
        if event.GetDirection():
            # Select next pane
            if selected == maxItems - 1:
                itemToSelect = 0
            else:
                itemToSelect = selected + 1
        else:
            # Previous pane
            if selected == 0:
                itemToSelect = maxItems - 1
            else:
                itemToSelect = selected - 1

        self._listBox.SetSelection(itemToSelect)
        self.HighlightPane()

    def PopulateListControl(self):
        """Populates the L{AuiPaneNavigator} with the panes in the AuiMgr"""
        self._panes = self._auimgr.GetAllPanes()
        names = [pane.name for pane in self._panes]
        self._listBox.AppendItems(sorted(names))

    def OnItemSelected(self, event):
        """Handles the wx.EVT_LISTBOX_DCLICK event"""
        self.CloseDialog()

    def CloseDialog(self):
        """Closes the L{AuiPaneNavigator} dialog"""
        self._selectedItem = self._listBox.GetStringSelection()
        self._auimgr.HideHint()
        self.EndModal(wx.ID_OK)

    def GetCloseKeys(self):
        """Get the list of keys that can dismiss the dialog
        @return: list of long (wx.WXK_*)

        """
        return self._close_keys

    def GetNavigationKeys(self):
        """Get the list of navigation key(s)
        @return: list of long (wx.WXK_*)

        """
        return self._navi_keys

    def GetSelection(self):
        """Get the index of the selected page"""
        return self._selectedItem

    def HighlightPane(self):
        """Highlight the currently selected pane"""
        sel = self._listBox.GetStringSelection()
        pane = self._auimgr.GetPane(sel)
        if pane.IsOk():
            self._auimgr.ShowHint(pane.window.GetScreenRect())
            # NOTE: this is odd but it is the only way for the focus to
            #       work correctly on wxMac...
            wx.CallAfter(self._listBox.SetFocus)
            self._listBox.SetFocus()

    def SetCloseKeys(self, keylist):
        """Set the keys that can be used to dismiss the L{AuiPaneNavigator}
        window.
        @param keylist: list of key codes

        """
        self._close_keys = keylist

    def SetNavigationKeys(self, keylist):
        """Set the key(s) to advance the selection in the pane list
        @param keylist: list of key codes

        """
        self._navi_keys = keylist

    def ShowModal(self):
        # Set focus on the list box to avoid having to click on it to change
        # the tab selection under GTK.
        self._listBox.SetFocus()
        self._listBox.SetSelection(0)
        return super(AuiPaneNavigator, self).ShowModal()


#!/usr/bin/env python
###############################################################################
# Name: Editra.py                                                             #
# Purpose: Implements Editras App object and the Main method                  #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
This module defines the Editra Application object and the Main method for
running Editra.

@summary: Editra's main application object and MainLoop

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: Editra.py 71718 2012-06-12 13:25:48Z CJP $"
__revision__ = "$Revision: 71718 $"

#--------------------------------------------------------------------------#
# Dependencies
import os
import sys

# Due to some methods that were added in 2.8.3 being used in a large number
# of places Editra has become incompatible with wxPython 2.8.1.1 and earlier.
# So ensure correct version of wxPython can be loaded
if not hasattr(sys, 'frozen') and 'wx' not in sys.modules:
    import wxversion
    wxversion.ensureMinimal('2.8')

import codecs
import base64
import locale
import getopt
import shutil
import wx

# The event handler mixin is now part of wxPython proper, but there hasn't
# been an official release with it yet, so try to import the official module
# but fallback to our own copy if it fails.
try:
    import wx.lib.eventStack as events
except:
    import extern.events as events

# Try and import a system installed version of pkg_resources else fallback to
# the one bundled with Editra's source.
try:
    from pkg_resources import resource_filename
except ImportError:
    from extern.pkg_resources import resource_filename

# Editra Libraries
import ed_glob
import ed_i18n
import profiler
import util
import dev_tool
import ed_main
import ed_art
import ed_txt
import ed_event
import updater
import plugin
import ed_ipc
import ed_session
import ebmlib
from syntax import synglob

#--------------------------------------------------------------------------#
# Global Variables
ID_UPDATE_CHECK = wx.NewId()

_ = wx.GetTranslation
#--------------------------------------------------------------------------#

class Editra(wx.App, events.AppEventHandlerMixin):
    """The Editra Application Object
    @deprecated: L{GetMainWindow}

    """
    def __init__(self, *args, **kargs):
        """Initialize that main app and its attributes
        @postcondition: application is created and ready to be run in mainloop

        """
        wx.App.__init__(self, *args, **kargs)
        events.AppEventHandlerMixin.__init__(self)

        # Attributes
        self._log = dev_tool.DEBUGP
        self._lock = False
        self._windows = dict()
        self._isexiting = False

        # Disable debug popups
        wx.Log.EnableLogging(False)
        # XXX: Temporary - disable assertions on OSX to work around 
        #      upstream bug in drawing code "couldnt draw the rotated text"
        if wx.Platform == '__WXMAC__':
            self.SetAssertMode(wx.PYAPP_ASSERT_SUPPRESS)

        # Purge old logs
        logfile = dev_tool.EdLogFile()
        logfile.PurgeOldLogs(7)

        if ed_glob.SINGLE:
            # Setup the instance checker
            instance_name = u"%s-%s" % (self.GetAppName(), wx.GetUserId())
            lockpath = wx.StandardPaths.Get().GetTempDir()
            self._instance = wx.SingleInstanceChecker(instance_name, path=lockpath)
            if self._instance.IsAnotherRunning():
                try:
                    opts, args = ProcessCommandLine()
                except getopt.GetoptError, msg:
                    self._log("[app][err] %s" % str(msg))
                    args = list()
                    opts = dict()

                exml = ed_ipc.IPCCommand()
                if len(args):
                    nargs = list()
                    for p in args:
                        try:
                            p = ebmlib.GetAbsPath(p)
                        except:
                            pass
                        fxml = ed_ipc.IPCFile()
                        fxml.value = p
                        nargs.append(fxml)
                    exml.filelist = nargs
                arglist = list()
                for arg, val in opts.items():
                    axml = ed_ipc.IPCArg()
                    axml.name = arg
                    axml.value = val
                    arglist.append(axml)
                exml.arglist = arglist

                # TODO: need to process other command line options as well i.e) -g
                self._log("[app][info] Sending: %s" % exml.Xml)
                key = profiler.Profile_Get('SESSION_KEY')
                if ebmlib.IsUnicode(key):
                    key = key.encode(sys.getfilesystemencoding(), 'replace')
                rval = ed_ipc.SendCommands(exml, key)
                # If sending the command failed then let the editor startup
                # a new instance
                if not rval:
                    self._isfirst = True
            else:
                self._log("[app][info] Starting Ipc server...")
                # Set the session key and save it to the users profile so
                # that other instances can access the server
                key = base64.b64encode(os.urandom(8), 'zZ')
                uname = wx.GetUserName()
                if ebmlib.IsUnicode(uname):
                    uname = uname.encode(sys.getfilesystemencoding(), 'replace')
                key = uname + key
                profiler.Profile_Set('SESSION_KEY', key)
                profiler.Profile_Set('ISBINARY', hasattr(sys, 'frozen'))
                path = profiler.Profile_Get('MYPROFILE')
                profiler.TheProfile.Write(path)
                try:
                    self._server = ed_ipc.EdIpcServer(self, key)
                    self._server.start()
                except Exception, msg:
                    self._log("[app][err] Failed to start ipc server")
                    self._log("[app][err] %s" % str(msg))
                    self._server = None
                self._isfirst = True
        else:
            self._isfirst = True

        # Setup Plugins after locale as they may have resource that need to
        # be loaded.
        if self._isfirst:
            self._pluginmgr = plugin.PluginManager()

            self._log("[app][info] Registering Editra's ArtProvider")
            wx.ArtProvider.PushProvider(ed_art.EditraArt())

        # Check if libenchant has been loaded or need to be
        import extern.stcspellcheck as stcspellcheck
        checker = stcspellcheck.STCSpellCheck
        if not checker.isEnchantOk():
            spref = profiler.Profile_Get('SPELLCHECK', default=dict())
            libpath = spref.get('epath', u'')
            checker.reloadEnchant(libpath)
            # TODO: log if load fails here

    def AddMessageCatalog(self, name, path):
        """Add a catalog lookup path to the app
        @param name: name of catalog (i.e 'projects')
        @param path: catalog lookup path

        """
        if self.locale is not None:
            path = resource_filename(path, 'locale')
            self.locale.AddCatalogLookupPathPrefix(path)
            self.locale.AddCatalog(name)

    def OnInit(self):
        """Initialize the Editor
        @note: this gets called before __init__
        @postcondition: custom artprovider and plugins are loaded

        """
        self.SetAppName(ed_glob.PROG_NAME)

        self._log = dev_tool.DEBUGP
        self._log("[app][info] Editra is Initializing")

        # Load user preferences
        self.profile_updated = InitConfig()
        self._isfirst = False # Is the first instance
        self._instance = None

        # Setup Locale
        locale.setlocale(locale.LC_ALL, '')
        langId = ed_i18n.GetLangId(profiler.Profile_Get('LANG'))
        if wx.Locale.IsAvailable(langId):
            self.locale = wx.Locale(langId)
            if self.locale.GetCanonicalName() in ed_i18n.GetAvailLocales():
                self._log("[app][info] Loaded Locale '%s'" % self.locale.CanonicalName)
                self.locale.AddCatalogLookupPathPrefix(ed_glob.CONFIG['LANG_DIR'])
                self.locale.AddCatalog(ed_glob.PROG_NAME)
            else:
                self._log("[app][err] Unknown Locale '%s'" % self.locale.CanonicalName)
                del self.locale
                self.locale = None
        else:
            self._log("[app][err] The locale %s is not available!" %  profiler.Profile_Get('LANG'))
            self.locale = None

        # Check and set encoding if necessary
        d_enc = profiler.Profile_Get('ENCODING')
        if not d_enc:
            profiler.Profile_Set('ENCODING', ed_txt.DEFAULT_ENCODING)
        else:
            # Ensure the default encoding is valid
            # Fixes up older installs on some systems that may have an
            # invalid encoding set.
            try:
                codecs.lookup(d_enc)
            except (LookupError, TypeError):
                self._log("[app][err] Resetting bad encoding: %s" % d_enc)
                profiler.Profile_Set('ENCODING', ed_txt.DEFAULT_ENCODING)

        # Setup the Error Reporter
        if profiler.Profile_Get('REPORTER', 'bool', True):
            sys.excepthook = dev_tool.ExceptionHook

        #---- Bind Events ----#
        self.Bind(wx.EVT_ACTIVATE_APP, self.OnActivate)
        self.Bind(wx.EVT_MENU, self.OnNewWindow, id=ed_glob.ID_NEW_WINDOW)
        self.Bind(wx.EVT_MENU, self.OnCloseWindow)
        self.Bind(ed_event.EVT_NOTIFY, self.OnNotify)
        self.Bind(ed_ipc.EVT_COMMAND_RECV, self.OnCommandReceived)

        # Splash a warning if version is not a final version
        if profiler.Profile_Get('APPSPLASH'):
            import edimage
            splash_img = edimage.splashwarn.GetBitmap()
            self.splash = wx.SplashScreen(splash_img, wx.SPLASH_CENTRE_ON_PARENT | \
                                          wx.SPLASH_NO_TIMEOUT, 0, None, wx.ID_ANY)
            self.splash.Show()

        return True

    def Destroy(self):
        """Destroy the application"""
        try:
            # Cleanup the instance checker
            del self._instance
        except AttributeError:
            pass
        wx.App.Destroy(self)

    def DestroySplash(self):
        """Destroy the splash screen"""
        # If is created and not dead already
        if getattr(self, 'splash', None) is not None and \
           isinstance(self.splash, wx.SplashScreen):
            self.splash.Destroy()
            self.splash = None

    def Exit(self, force=False):
        """Exit the program
        @postcondition: If no toplevel windows are present program will exit.
        @postcondition: Program may remain open if an open window is locking.

        """
        self._isexiting = True
        self._pluginmgr.WritePluginConfig()
        profiler.TheProfile.Write(profiler.Profile_Get('MYPROFILE'))
        if not self._lock or force:
            if getattr(self, '_server', None):
                self._server.Shutdown()

            try:
                # Cleanup the instance checker
                del self._instance
            except AttributeError:
                pass

            # Exit the app
            wx.App.ExitMainLoop(self)

    def GetLocaleObject(self):
        """Get the locale object owned by this app. Use this method to add
        extra catalogs for lookup.
        @return: wx.Locale or None

        """
        return self.locale

    def GetLog(self):
        """Returns the logging function used by the app
        @return: the logging function of this program instance

        """
        return self._log

    def GetMainWindow(self):
        """Returns reference to the instance of the MainWindow
        that is running if available, and None if not.
        @return: the L{MainWindow} of this app if it is open

        """
        self._log("[app][warn] Editra::GetMainWindow is deprecated")
        for window in self._windows:
            if isinstance(self._windows[window][0], ed_main.MainWindow):
                return self._windows[window][0]
        return None

    def GetActiveWindow(self):
        """Returns the active main window if there is one else it will
        just return some main window or none if there are no main windows
        @return: frame instance or None

        """
        awin = None
        for win in self.GetMainWindows():
            if win.IsActive():
                awin = win
                break

        if awin is None:
            awin = self.GetTopWindow()
            if not isinstance(awin, ed_main.MainWindow):
                if len(self.GetMainWindows()):
                    awin = self.GetMainWindows()[0]

        return awin

    def GetCurrentBuffer(self):
        """Get the current buffer from the active window or None
        @return: EditraStc

        """
        win = self.GetTopWindow()
        if not isinstance(win, ed_main.MainWindow):
            win = self.GetActiveWindow()
            if win is None:
                return None # UI dead?

        if isinstance(win, ed_main.MainWindow):
            nbook = win.GetNotebook()
            if nbook:
                return nbook.GetCurrentCtrl()
        return None

    def GetMainWindows(self):
        """Returns a list of all open main windows
        @return: list of L{MainWindow} instances of this app (list may be empty)

        """
        mainw = list()
        for window in self._windows:
            try:
                if isinstance(self._windows[window][0], ed_main.MainWindow):
                    mainw.append(self._windows[window][0])
            except AttributeError:
                continue
        return mainw

    def GetOpenWindows(self):
        """Returns a list of open windows
        @return: list of all open windows owned by app

        """
        return self._windows

    def GetPluginManager(self):
        """Returns the plugin manager used by this application
        @return: Apps plugin manager
        @see: L{plugin}

        """
        return self._pluginmgr

    def GetProfileUpdated(self):
        """Was the profile updated 
        @return: bool

        """
        return self.profile_updated

    def GetWindowInstance(self, wintype):
        """Get an instance of an open window if one exists
        @param wintype: Class type of window to look for
        @precondition: Window must have called L{RegisterWindow}
        @return: Instance of window or None

        """
        for win in self._windows:
            if isinstance(self._windows[win][0], wintype):
                return self._windows[win][0]
        return None

    def IsLocked(self):
        """Returns whether the application is locked or not
        @return: whether a window has locked the app from closing or not

        """
        return self._lock

    def IsOnlyInstance(self):
        """Check if this app is the the first instance that is running
        @return: bool

        """
        return self._isfirst

    def Lock(self):
        """Locks the app from exiting
        @postcondition: program is locked from exiting

        """
        self._lock = True

    def OpenFile(self, filename, line=-1):
        """Open a file in the currently active window
        @param filename: file path
        @keyword line: int

        """
        window = self.GetTopWindow()
        if not isinstance(window, ed_main.MainWindow):
            window = None

        try:
            encoding = sys.getfilesystemencoding()
            fname = ed_txt.DecodeString(filename, encoding)

            if profiler.Profile_Get('OPEN_NW', default=False):
                self.OpenNewWindow(fname, window)
            elif window:
                window.DoOpen(ed_glob.ID_COMMAND_LINE_OPEN, fname, line)

                # Make sure the window is brought to the front
                if window.IsIconized():
                    window.Iconize(False)
                window.Raise()
            else:
                # Some unlikely error condition
                self._log("[app][err] OpenFile unknown error: %s" % filename)
                
        except Exception, msg:
            self._log("[app][err] Failed to open file: %s" % str(msg))

    def MacNewFile(self):
        """Stub for future use"""
        pass

    def MacOpenFile(self, filename):
        """Macintosh Specific code for opening files that are associated
        with the editor and double clicked on after the editor is already
        running.
        @param filename: file path string
        @postcondition: if L{MainWindow} is open file will be opened in notebook

        """
        self._log("[app][info] MacOpenFile Fired")
        self.OpenFile(filename, line=-1)

    def MacPrintFile(self, filename):
        """Stub for future use
        @param filename: file to print

        """
        pass

    def MacReopenApp(self):
        """Handle kAEReopenApplication when dock icons is clicked on"""
        frame = self.GetTopWindow()
        if frame is not None:
            if frame.IsIconized():
                frame.Iconize(False)
            frame.Raise()

    def OnActivate(self, evt):
        """Activation Event Handler
        @param evt: wx.ActivateEvent

        """
        if evt.GetActive():
            self._log("[app][info] I'm Awake!!")
            # Refresh Clipboard Ring
            ed_main.MainWindow.UpdateClipboardRing()

#            frame = self.GetTopWindow()
#            if frame is not None:
#                if frame.IsIconized():
#                    frame.Iconize(False)
#                frame.Raise()
        else:
            self._log("[app][info] Going to sleep")
        evt.Skip()

    def OnExit(self, evt=None, force=False):
        """Handle application exit request
        @keyword evt: event that called this handler
        @keyword force: Force an exit

        """
        e_id = -1
        if evt:
            e_id = evt.GetId()

        if e_id == ed_glob.ID_EXIT:
            self._isexiting = True
            # First loop is to ensure current top window is
            # closed first
            for win in self.GetMainWindows():
                if win.IsActive():
                    result = win.Close()
                    if result:
                        self._isexiting = False
                        break
                    return
            for win in self.GetMainWindows():
                win.Raise()
                result = win.Close()
                if not result:
                    self._isexiting = False
                    break
            self.Exit(force)
        else:
            if evt:
                evt.Skip()

    def OnNewWindow(self, evt):
        """Create a new editing window
        @param evt: wx.EVT_MENU

        """
        if evt.GetId() == ed_glob.ID_NEW_WINDOW:
            frame = evt.GetEventObject().GetMenuBar().GetFrame()
            self.OpenNewWindow(caller=frame)
        else:
            evt.Skip()

    def OnCommandReceived(self, evt):
        """Receive commands from the IPC server
        @todo: move command processing into own module

        """
        # Guard against events that come in after shutdown
        # from server thread.
        if not self or self._isexiting or not self.IsMainLoopRunning():
            return

        self._log("[app][info] IN OnCommandReceived")
        cmds = evt.GetCommands()
        if isinstance(cmds, ed_ipc.IPCCommand):
            self._log("[app][info] OnCommandReceived %s" % cmds.Xml)
            if not len(cmds.filelist):
                self.OpenNewWindow()
            else:
                # TODO: change goto line handling to require one
                #       arg per file specified on the command line
                #       i.e) -g 23,44,100
                line = -1
                for argobj in cmds.arglist:
                    arg = argobj.name
                    if arg == '-g':
                        line = int(argobj.value)
                        if line > 0:
                            line -= 1
                        break

                for fname in cmds.filelist:
                    self.OpenFile(fname.value, line)

    def OnCloseWindow(self, evt):
        """Close the currently active window
        @param evt: wx.MenuEvent

        """
        if evt.GetId() in [ed_glob.ID_CLOSE, ed_glob.ID_CLOSE_WINDOW]:
            for window in wx.GetTopLevelWindows():
                if hasattr(window, 'IsActive') and window.IsActive():
                    if hasattr(window, 'Close'):
                        window.Close()
                    break
        else:
            evt.Skip()

    def OpenNewWindow(self, fname=u'', caller=None):
        """Open a new window
        @keyword fname: Open a file in the new window
        @keyword caller: MainWindow that called to open this one
        @return: the new window

        """
        frame = ed_main.MainWindow(None, wx.ID_ANY,
                                   profiler.Profile_Get('WSIZE'),
                                   ed_glob.PROG_NAME)
        if caller:
            pos = caller.GetPosition()
            frame.SetPosition((pos.x + 22, pos.y + 22))

        self.RegisterWindow(repr(frame), frame, True)
        self.SetTopWindow(frame)
        if isinstance(fname, basestring) and fname != u'':
            frame.DoOpen(ed_glob.ID_COMMAND_LINE_OPEN, fname)
        frame.Show(True)

        # Ensure frame gets an Activate event when shown
        # this doesn't happen automatically on windows
        if wx.Platform == '__WXMSW__':
            wx.PostEvent(frame, wx.ActivateEvent(wx.wxEVT_ACTIVATE, True))
        return frame

    def OnNotify(self, evt):
        """Handle notification events
        @param evt: L{ed_event.NotificationEvent}

        """
        e_val = evt.GetValue()
        if evt.GetId() == ID_UPDATE_CHECK and \
           isinstance(e_val, tuple) and e_val[0]:
            self.DestroySplash()
            mdlg = wx.MessageDialog(self.GetActiveWindow(),
                                    _("An updated version of Editra is available\n"
                                      "Would you like to download Editra %s now?") %\
                                      e_val[1], _("Update Available"),
                                    wx.YES_NO|wx.YES_DEFAULT|wx.CENTER|wx.ICON_INFORMATION)
            if mdlg.ShowModal() == wx.ID_YES:
                dl_dlg = updater.DownloadDialog(None, wx.ID_ANY,
                                                _("Downloading Update"))
                dp_sz = wx.GetDisplaySize()
                dl_dlg.SetPosition(((dp_sz[0] - (dl_dlg.GetSize()[0] + 5)), 25))
                dl_dlg.Show()
            mdlg.Destroy()
        else:
            evt.Skip()

    def RegisterWindow(self, name, window, can_lock=False):
        """Registers winows with the app. The name should be the
        repr of window. The can_lock parameter is a boolean stating
        whether the window can keep the main app running after the
        main frame has exited.
        @param name: name of window
        @param window: reference to window object
        @keyword can_lock: whether window can lock exit or not

        """
        self._windows[name] = (window, can_lock)

    def ReloadArtProvider(self):
        """Reloads the custom art provider onto the artprovider stack
        @postcondition: artprovider is removed and reloaded

        """
        try:
            wx.ArtProvider.PopProvider()
        finally:
            wx.ArtProvider.PushProvider(ed_art.EditraArt())

    def UnLock(self):
        """Unlocks the application
        @postcondition: application is unlocked so it can exit

        """
        self._lock = False

    def UnRegisterWindow(self, name):
        """Unregisters a named window with the app if the window
        was the top window and if other windows that can lock are
        registered in the window stack it will promote the next one
        it finds to be the top window. If no windows that fit this
        criteria are found it will close the application.
        @param name: name of window to unregister

        """
        if name in self._windows:
            self._windows.pop(name)
            
            if not len(self._windows):
                self._log("[app][info] No more open windows shutting down")
                self.Exit()
                return

            # TODO: WXBUG? calling GetTopWindow when there are no more top
            #       level windows causes a crash under MSW. Moving this line
            #       above the previous check can reproduce the error.
            cur_top = self.GetTopWindow()
            if name == repr(cur_top):
                found = False
                for key in self._windows:
                    if self._windows[key][1]:
                        self._log("[app][info] Promoting %s to top" % key)
                        try:
                            self.SetTopWindow(self._windows[key][0])
                        except Exception:
                            continue
                        found = True
                        break

                if not found:
                    self._log("[app][info] No more top windows exiting app")
                    self.UnLock()
                    self.Exit()
            else:
                self._log("[app][info] UnRegistered %s" % name)
        else:
            self._log("[app][warn] The window %s is not registered" % name)

    def WindowCanLock(self, winname):
        """Checks if a named window can lock the application or
        not. The window must have been previously registered with
        a call to RegisterWindow for this function to have any
        real usefullness.
        @param winname: name of window to query

        """
        if winname in self._windows:
            return self._windows[winname][1]
        else:
            self._log("[app][warn] the window %s has "
                      "not been registered" % winname)
            return False

#--------------------------------------------------------------------------#

def InitConfig():
    """Initializes the configuration data
    @postcondition: all configuration data is set

    """
    # Check if a custom config directory was specified on the commandline
    if ed_glob.CONFIG['CONFIG_BASE'] is not None:
        # TODO: there is a bug when the first time the config is created
        #       where the settings will not be saved until second launching.
        config_base = os.path.abspath(ed_glob.CONFIG['CONFIG_BASE'])
    else:
        # Look for a profile directory on the system level. If this directory
        # exists Use it instead of the user one. This will allow for running
        # Editra from a portable drive or for system administrators to enforce
        # settings on a system installed version.
        config_base = util.ResolvConfigDir(u'.Editra', True) 

    if os.path.exists(config_base):
        ed_glob.CONFIG['CONFIG_BASE'] = config_base
        ed_glob.CONFIG['PROFILE_DIR'] = os.path.join(config_base, u"profiles")
        ed_glob.CONFIG['PROFILE_DIR'] += os.sep
        ed_glob.CONFIG['ISLOCAL'] = True
    else:
        config_base = wx.StandardPaths.Get().GetUserDataDir()
        ed_glob.CONFIG['PROFILE_DIR'] = util.ResolvConfigDir(u"profiles")

    # Check for if config directory exists and if profile is from the current
    # running version of Editra.
    profile_updated = False
    if util.HasConfigDir() and os.path.exists(ed_glob.CONFIG['PROFILE_DIR']):
        if profiler.ProfileIsCurrent():
            pstr = profiler.GetProfileStr()
            # If using local(portable) config the profile string is stored
            # as a relative path that just names the config file.
            if ed_glob.CONFIG['ISLOCAL']:
                pstr = os.path.join(ed_glob.CONFIG['PROFILE_DIR'], pstr)
            pstr = util.RepairConfigState(pstr)
            dev_tool.DEBUGP("[InitConfig][info] Loading profile: %s" % repr(pstr))
            profiler.TheProfile.Load(pstr)
        else:
            dev_tool.DEBUGP("[InitConfig][info] Updating Profile to current version")

            # When upgrading from an older version make sure all
            # config directories are available.
            for cfg in ("cache", "styles", "plugins", "profiles", "sessions"):
                if not util.HasConfigDir(cfg):
                    util.MakeConfigDir(cfg)

            # Load and update profile
            pstr = profiler.GetProfileStr()
            pstr = util.RepairConfigState(pstr)
            profiler.TheProfile.Load(pstr)
            profiler.TheProfile.Update()

            #---- Temporary Profile Adaption ----#

            # Added after 0.5.32
            mconfig = profiler.Profile_Get('LEXERMENU', default=None)
            if mconfig is None:
                mconfig = [ synglob.LANG_C, synglob.LANG_CPP,
                            synglob.LANG_BASH, synglob.LANG_CSS,
                            synglob.LANG_HTML, synglob.LANG_JAVA,
                            synglob.LANG_LISP, synglob.LANG_PERL,
                            synglob.LANG_PHP, synglob.LANG_PYTHON,
                            synglob.LANG_RUBY, synglob.LANG_SQL,
                            synglob.LANG_XML]
                mconfig.sort()
                profiler.Profile_Set('LEXERMENU', mconfig)

            # GUI_DEBUG mode removed in 0.2.5
            mode = profiler.Profile_Get('MODE')
            if mode == 'GUI_DEBUG':
                profiler.Profile_Set('MODE', 'DEBUG')

            # This key has been removed so clean it from old profiles
            profiler.Profile_Del('LASTCHECK')

            # Print modes don't use strings anymore
            if isinstance(profiler.Profile_Get('PRINT_MODE'), basestring):
                profiler.Profile_Set('PRINT_MODE', ed_glob.PRINT_BLACK_WHITE)

            # Simplifications to eol mode persistence (0.4.28)
            # Keep for now till plugins are updated
            #profiler.Profile_Del('EOL') # changed to EOL_MODE

            # After 0.4.65 LAST_SESSION now points a session file and not
            # to a list of files to open.
            ed_glob.CONFIG['SESSION_DIR'] = util.ResolvConfigDir(u"sessions")
            smgr = ed_session.EdSessionMgr()
            sess = profiler.Profile_Get('LAST_SESSION')
            if isinstance(sess, list) or not sess:
                profiler.Profile_Set('LAST_SESSION', smgr.DefaultSession)
            else:
                # After 0.6.58 session is reduced to a name instead of path
                if sess and os.path.sep in sess:
                    name = smgr.SessionNameFromPath(sess)
                    profiler.Profile_Set('LAST_SESSION', name)

            #---- End Temporary Profile Adaption ----#

            # Write out updated profile
            profiler.TheProfile.Write(pstr)

            profile_updated = True
    else:
        # Fresh install
        util.CreateConfigDir()

        # Check and upgrade installs from old location
        success = True
        try:
            success = UpgradeOldInstall()
        except Exception, msg:
            dev_tool.DEBUGP("[InitConfig][err] %s" % msg)
            success = False

        if not success:
            old_cdir = u"%s%s.%s%s" % (wx.GetHomeDir(), os.sep,
                                       ed_glob.PROG_NAME, os.sep)
            msg = ("Failed to upgrade your old installation\n"
                   "To retain your old settings you may need to copy some files:\n"
                   "\nFrom: %s\n\nTo: %s") % (old_cdir, config_base)
            wx.MessageBox(msg, "Upgrade Failed", style=wx.ICON_WARNING|wx.OK)

        # Set default eol for windows
        if wx.Platform == '__WXMSW__':
            profiler.Profile_Set('EOL_MODE', ed_glob.EOL_MODE_CRLF)
            profiler.Profile_Set('ICONSZ', (16, 16))
        elif wx.Platform == '__WXMAC__':
            # Default to 32x32 toolbar icons on OSX
            profiler.Profile_Set('ICONSZ', (32, 32))

    #---- Profile Loaded / Installed ----#

    # Set debug mode
    emode = profiler.Profile_Get('MODE')
    if 'DEBUG' in emode:
        ed_glob.DEBUG = True
        if emode.startswith('VERBOSE'):
            ed_glob.VDEBUG = True

    # Resolve resource locations
    ed_glob.CONFIG['CONFIG_DIR'] = util.ResolvConfigDir(u"")
    ed_glob.CONFIG['INSTALL_DIR'] = util.ResolvConfigDir(u"", True)
    ed_glob.CONFIG['KEYPROF_DIR'] = util.ResolvConfigDir(u"ekeys", True)
    ed_glob.CONFIG['SYSPIX_DIR'] = util.ResolvConfigDir(u"pixmaps", True)
    ed_glob.CONFIG['PLUGIN_DIR'] = util.ResolvConfigDir(u"plugins")
    ed_glob.CONFIG['THEME_DIR'] = util.ResolvConfigDir(os.path.join(u"pixmaps", u"theme"))
    ed_glob.CONFIG['LANG_DIR'] = util.ResolvConfigDir(u"locale", True)
    ed_glob.CONFIG['STYLES_DIR'] = util.ResolvConfigDir(u"styles")
    ed_glob.CONFIG['SYS_PLUGIN_DIR'] = util.ResolvConfigDir(u"plugins", True)
    ed_glob.CONFIG['SYS_STYLES_DIR'] = util.ResolvConfigDir(u"styles", True)
    ed_glob.CONFIG['TEST_DIR'] = util.ResolvConfigDir(os.path.join(u"tests", u"syntax"), True)

    # Make sure all standard config directories are there
    for cfg in ("cache", "styles", "plugins", "profiles", "sessions"):
        if not util.HasConfigDir(cfg):
            util.MakeConfigDir(cfg)
    ed_glob.CONFIG['CACHE_DIR'] = util.ResolvConfigDir(u"cache")
    ed_glob.CONFIG['SESSION_DIR'] = util.ResolvConfigDir(u"sessions")

    return profile_updated

#--------------------------------------------------------------------------#

def UpgradeOldInstall():
    """Upgrade an old installation and transfer all files if they exist
    @note: FOR INTERNAL USE ONLY
    @return: bool (True if success, False if failure)

    """
    old_cdir = u"%s%s.%s%s" % (wx.GetHomeDir(), os.sep,
                               ed_glob.PROG_NAME, os.sep)
    base = ed_glob.CONFIG['CONFIG_BASE']
    if base is None:
        base = wx.StandardPaths.Get().GetUserDataDir() + os.sep

    err = 0
    if os.path.exists(old_cdir) and \
       base.lower().rstrip(os.sep) != old_cdir.lower().rstrip(os.sep):
        for item in os.listdir(old_cdir):
            try:
                dest = os.path.join(base, item)
                item = os.path.join(old_cdir, item)
                if os.path.exists(dest):
                    if os.path.isdir(dest):
                        shutil.rmtree(dest, True)
                    else:
                        os.remove(dest)

                shutil.move(item, dest)
            except Exception, msg:
                util.Log("[Upgrade][err] %s" % msg)
                err += 1
                continue

        os.rmdir(old_cdir)

        # Load the copied over profile
        pstr = profiler.GetProfileStr()
        prof = os.path.basename(pstr)
        pstr = os.path.join(base, u"profiles", prof)
        if os.path.exists(pstr):
            profiler.TheProfile.Load(pstr)
            profiler.TheProfile.Update()
            profiler.UpdateProfileLoader()

        if not err:
            wx.MessageBox(_("Your profile has been updated to the latest "
                "version") + u"\n" + \
              _("Please check the preferences dialog to check "
                "your preferences"),
              _("Profile Updated"))

    return not err

#--------------------------------------------------------------------------#

def PrintHelp(err=None):
    """Print command line help
    @postcondition: Help is printed and program exits

    """
    if err is not None:
        sys.stderr.write(err + os.linesep)

    print(("Editra - %s - Developers Text Editor\n"
       "Cody Precord (2005-2012)\n\n"
       "usage: Editra [arguments] [files... ]\n\n"
       "Short Arguments:\n"
       "  -c    Set custom configuration directory at runtime\n"
       "  -d    Turn on console debugging (-dd for verbose debug)\n"
       "  -D    Turn off console debugging (overrides preferences)\n"
       "  -g    Open file to line (i.e Editra -g 10 file.txt)\n"
       "  -h    Show this help message\n"
       "  -p    Run Editra in the profiler (outputs to editra.prof).\n"
       "  -v    Print version number and exit\n"
       "  -S    Disable single instance checker\n"
       "\nLong Arguments:\n"
       "  --confdir arg     Set custom configuration directory at runtime\n"
       "  --debug           Turn on console debugging\n"
       "  --help            Show this help message\n"
       "  --auth            Print the ipc server info\n"
       "  --version         Print version number and exit\n"
       "  --profileOut arg  Run Editra in the profiler (arg is output file)\n"
      ) % ed_glob.VERSION)

    if err is None:
        os._exit(0)
    else:
        os._exit(1)

#--------------------------------------------------------------------------#

def ProcessCommandLine():
    """Process the command line switches
    @return: tuple ({switches,}, [args,])

    """
    try:
        items, args = getopt.getopt(sys.argv[1:], "dg:hp:vDSc:",
                                   ['debug', 'help', 'version', 'auth',
                                    'confdir=', 'profileOut='])
    except getopt.GetoptError, msg:
        # Raise error to console and exit
        PrintHelp(str(msg))
    
    # Process command line options
    opts = dict(items)
    for opt, value in dict(opts).items():
        if opt in ['-h', '--help']:
            PrintHelp()
        elif opt in ['-v', '--version']:
            print(ed_glob.VERSION)
            os._exit(0)
        elif opt in ['-d', '--debug'] and '-D' not in opts.keys():
            # If the debug flag is set more than once go into verbose mode
            if ed_glob.DEBUG:
                ed_glob.VDEBUG = True
            ed_glob.DEBUG = True
            opts.pop(opt)
        elif opt == '-D':
            ed_glob.DEBUG = False
            ed_glob.VDEBUG = False
            opts.pop('-D')
        elif opt == '-S':
            # Disable single instance checker
            ed_glob.SINGLE = False
            opts.pop(opt)
        elif opt in ['-c', '--confdir']:
            ed_glob.CONFIG['CONFIG_BASE'] = value
            opts.pop(opt)
        elif opt == '--profileOut':
            opts['-p'] = value
            opts.pop('--profileOut')
        elif opt == '-g':
            # Validate argument passed to -g
            if not value.isdigit():
                PrintHelp("error: -g requires a number as an argument!")
        else:
            pass

    # Return any unprocessed arguments
    return opts, args

#--------------------------------------------------------------------------#

def Main():
    """Configures and Runs an instance of Editra
    @summary: Parses command line options, loads the user profile, creates
              an instance of Editra and starts the main loop.

    """
    opts, args = ProcessCommandLine()

    if '-p' in opts:
        p_file = opts['-p']
        opts.pop('-p')

        if not len(p_file):
            # Fall back to default output file
            p_file = "editra.prof"
            
        import hotshot
        prof = hotshot.Profile(p_file)
        prof.runcall(_Main, opts, args)
        prof.close()
    else:
        _Main(opts, args)

def _Main(opts, args):
    """Main method
    @param opts: Commandline options
    @param args: Commandline arguments

    """
    # Put extern subpackage on path so that bundled external dependencies
    # can be found if needed.
    if not hasattr(sys, 'frozen'):
        epath = os.path.join(os.path.dirname(__file__), 'extern')
        if os.path.exists(epath):
            sys.path.append(epath)

    # Create Application
    dev_tool.DEBUGP("[main][app] Initializing application...")
    editra_app = Editra(False)

    # Print ipc server authentication info
    if '--auth' in opts:
        opts.pop('--auth')
        print "port=%d,key=%s" % (ed_ipc.EDPORT,
                                  profiler.Profile_Get('SESSION_KEY'))

    # Check if this is the only instance, if its not exit since
    # any of the opening commands have already been passed to the
    # master instance
    if not editra_app.IsOnlyInstance():
        dev_tool.DEBUGP("[main][info] Second instance exiting...")
        editra_app.Destroy()
        os._exit(0)

    # Set the timeout on destroying the splash screen
    wx.CallLater(2300, editra_app.DestroySplash)

    if profiler.Profile_Get('SET_WSIZE'):
        wsize = profiler.Profile_Get('WSIZE')
    else:
        wsize = (700, 450)
    frame = ed_main.MainWindow(None, wx.ID_ANY, wsize, ed_glob.PROG_NAME)
    frame.Maximize(profiler.Profile_Get('MAXIMIZED'))
    editra_app.RegisterWindow(repr(frame), frame, True)
    editra_app.SetTopWindow(frame)
    frame.Show(True)

    # Load Session Data
    # But not if there are command line args for files to open
    if profiler.Profile_Get('SAVE_SESSION', 'bool', False) and not len(args):
        smgr = ed_session.EdSessionMgr()
        session = profiler.Profile_Get('LAST_SESSION', default=u'')
        if isinstance(session, list):
            # Check for format conversion from previous versions
            profiler.Profile_Set('LAST_SESSION', smgr.DefaultSession)
            session = smgr.DefaultSession
        frame.GetNotebook().LoadSessionFile(session)
        del session

    # Unlike wxMac/wxGTK Windows doesn't post an activate event when a window
    # is first shown, so do it manually to make sure all event handlers get
    # pushed.
    if wx.Platform == '__WXMSW__':
        wx.PostEvent(frame, wx.ActivateEvent(wx.wxEVT_ACTIVATE, True))

    # Do update check if preferences say its ok to do so
    isadmin = os.access(ed_glob.CONFIG['INSTALL_DIR'], os.R_OK|os.W_OK)
    if isadmin and profiler.Profile_Get('CHECKUPDATE', default=True):
        uthread = updater.UpdateThread(editra_app, ID_UPDATE_CHECK)
        uthread.start()

    if len(args):
        line = -1
        if '-g' in opts:
            line = max(0, int(opts.pop('-g')) - 1)

        # TODO: should line arg only be applied to the first file name or all?
        #       currently apply to all.
        for arg in args:
            try:
                fname = ed_txt.DecodeString(arg, sys.getfilesystemencoding())
                fname = ebmlib.GetAbsPath(fname)
                frame.DoOpen(ed_glob.ID_COMMAND_LINE_OPEN, fname, line)
            except IndexError:
                dev_tool.DEBUGP("[main][err] IndexError on commandline args")

    # Notify that profile was updated
    if editra_app.GetProfileUpdated():
        editra_app.DestroySplash()
        # Make sure window iniliazes to default position
        profiler.Profile_Del('WPOS')
        wx.MessageBox(_("Your profile has been updated to the latest "
                        "version") + u"\n" + \
                      _("Please check the preferences dialog to verify "
                        "your preferences"),
                      _("Profile Updated"))

    # 3. Start Applications Main Loop
    dev_tool.DEBUGP("[main][info] Starting MainLoop...")
    wx.CallAfter(frame.Raise)

    # Install handlers to exit app if os is shutting down/restarting
    ebmlib.InstallTermHandler(editra_app.Exit, force=True)

    editra_app.MainLoop()
    dev_tool.DEBUGP("[main][info] MainLoop finished exiting application")
    os._exit(0)

#-----------------------------------------------------------------------------#
if __name__ == '__main__':
    Main()



###############################################################################
# Name: testECLUtil.py                                                        #
# Purpose: Unit tests for the eclib.eclutil module                            #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittest cases for testing the eclib.eclutil functions"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: testECLUtil.py 62952 2009-12-20 02:01:47Z CJP $"
__revision__ = "$Revision: 62952 $"

#-----------------------------------------------------------------------------#
# Imports
import wx
import unittest

# Local modules
import common

# Module to test
import eclib

#-----------------------------------------------------------------------------#
# Test Class

class ECLUtilTest(unittest.TestCase):
    """Tests for the eclib.eclutil functions"""
    def setUp(self):
        self.app = wx.App(False)

    def tearDown(self):
        self.app.Destroy()

    #---- Tests ----#

    def testAdjustAlpha(self):
        """Test adjusting the alpha channel of a colour"""
        c = wx.Colour(255, 255, 255, 255)

        c = eclib.AdjustAlpha(c, 128)
        self.assertEquals(c.Alpha(), 128, "Alpha == %d" % c.Alpha())

        c = eclib.AdjustAlpha(c, 0)
        self.assertEquals(c.Alpha(), 0)

        self.assertRaises(ValueError, eclib.AdjustAlpha, c, -3)

    def testAdjustColour(self):
        """Test that a valid colour results are returned"""
        c = wx.Colour(125, 125, 125)

        # Check that the color was brightened
        c2 = eclib.AdjustColour(c, 50)
        self.assertTrue(sum(c.Get()) < sum(c2.Get()),
                        "Failed to lighten colour")

        # Check that the color was darkened
        c2 = eclib.AdjustColour(c, -50)
        self.assertTrue(sum(c.Get()) > sum(c2.Get()),
                        "Failed to darken colour")

    def testBestLabelColour(self):
        """Test getting the best label colour"""
        c = eclib.BestLabelColour(wx.BLACK)
        self.assertEquals((255, 255, 255), c.Get())

        c = eclib.BestLabelColour(wx.WHITE)
        self.assertEquals((0, 0, 0), c.Get())

    def testHexToRGB(self):
        """Test conversion of a hex value to a rgb tuple"""
        c = eclib.HexToRGB("000000")
        self.assertEquals(c, [0, 0, 0])

        c = eclib.HexToRGB("#000000")
        self.assertEquals(c, [0, 0, 0])

        c = eclib.HexToRGB("#0000FF")
        self.assertEquals(c, [0, 0, 255])

        c = eclib.HexToRGB("#FF00FF")
        self.assertEquals(c, [255, 0, 255])

        self.assertRaises(IndexError, eclib.HexToRGB, u"")
        self.assertRaises(ValueError, eclib.HexToRGB, "FF23GG")



###############################################################################
# Name: diff.py                                                               #
# Purpose: Define Diff/Patch file syntax for highlighting and other features  #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: diff.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Diff/Patch files

"""

__author__ = "Cody Precord <cprecord@editra.org"
__svnid__ = "$Id: _diff.py 66108 2010-11-10 21:04:54Z CJP $"
__revision__ = "$Revision: 66108 $"

#-----------------------------------------------------------------------------#
# Imports
import wx
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
# None
#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [(stc.STC_DIFF_ADDED,    'global_style'),
                (stc.STC_DIFF_COMMAND,  'pre_style'),
                (stc.STC_DIFF_COMMENT,  'comment_style'),
                (stc.STC_DIFF_DEFAULT,  'default_style'),
                (stc.STC_DIFF_DELETED,  'error_style'),
                (stc.STC_DIFF_HEADER,   'comment_style'),
                (stc.STC_DIFF_POSITION, 'pre_style')]

if wx.VERSION >= (2, 9, 0, 0, ''):
    SYNTAX_ITEMS.append((stc.STC_DIFF_CHANGED,  'default_style')) #TODO

#---- Extra Properties ----#
FOLD = ('fold', '1')
FOLD_COMPACT = ('fold.compact', '1')

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Diff files""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_DIFF)

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD, FOLD_COMPACT]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return ['--- ']



###############################################################################
# Name: weblib.py                                                             #
# Purpose: Web an network utilties                                            #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2010 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Editra Buisness Model Library: Web Utilities

Utility functions for working with web and other networking protocols

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: e_weblib.py 66131 2010-11-13 05:22:48Z CJP $"
__revision__ = "$Revision: 66131 $"

__all__ = ['SOAP12Message',]

#-----------------------------------------------------------------------------#
# imports
import urllib2
import httplib

#-----------------------------------------------------------------------------#
_SOAP_TPL = """<?xml version=\"1.0\" encoding=\"utf-8\"?>
<soap12:Envelope xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:soap12=\"http://www.w3.org/2003/05/soap-envelope\">
<soap12:Body>
    %(msg)s
</soap12:Body>
</soap12:Envelope>

"""

_SM_TPL = """<?xml version="1.0" encoding="UTF-8"?>
<SOAP-ENV:Envelope 
SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"  
xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/">
<SOAP-ENV:Body>
    %(msg)s
</SOAP-ENV:Body>
</SOAP-ENV:Envelope>
"""

#-----------------------------------------------------------------------------#

class SOAP12Message(object):
    """Class for creating and sending a message
    using the SOAP protocol.

    """
    def __init__(self, host, request, msg, action=""):
        """Create the message object
        @param host: host the message will be sent to (url)
        @param request: POST request
        @param msg: XML Body text
        @keyword action: SoapAction

        """
        assert len(host), "Must specify a valid host"
        super(SOAP12Message, self).__init__()

        # Attributes
        self._host = host
        self._request = request
        self._msg = msg
        self._action = action
        self._http = httplib.HTTP(self._host, 80)

    @property
    def MessageBody(self):
        soapmsg = _SOAP_TPL % dict(msg=self._msg)
        soapmsg = soapmsg.replace("\n", "\r\n")
        return soapmsg

    def Send(self):
        """Send the message"""
        # Create the SOAP message
        soapmsg = self.MessageBody

        # Setup Headers
        self._http.putrequest("POST", self._request)
        self._http.putheader("Host", self._host)
#        self._http.putheader("User-Agent", "Python post")
        self._http.putheader("Content-Type", "application/soap+xml; charset=utf-8")
        self._http.putheader("Content-Length", "%d" % len(soapmsg))
        self._http.putheader("SOAPAction", '"%s"' % self._action)
        self._http.endheaders()

        # Send it
        self._http.send(soapmsg)

    def GetReply(self):
        """Get the reply (may block for a long time)
        @return: (statuscode, statusmessage, header)

        """
        return self._http.getreply()



###############################################################################
# Name: s.py                                                                  #
# Purpose: Define S and R syntax for highlighting and other features          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: s.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for the S and R statistical languages

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _s.py 70466 2012-01-26 20:55:16Z CJP $"
__revision__ = "$Revision: 70466 $"

#-----------------------------------------------------------------------------#
# Imports
from pygments.token import Token
from pygments.lexers import get_lexer_by_name
import wx
import wx.stc as stc

#Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#
# Style Id's

STC_S_DEFAULT, \
STC_S_COMMENT, \
STC_S_NUMBER, \
STC_S_STRING, \
STC_S_STRINGEOL, \
STC_S_OPERATOR, \
STC_S_KEYWORD = range(7)

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

KEYWORDS = ("all array break call else exp for function if length library list "
            "match max mean min while return try NULL NA TRUE FALSE")

R_KEYWORDS = "if else repeat while function for in next break TRUE FALSE NULL NA Inf NaN"
R_KEYWORDS2 = ("abbreviate abline abs acf acos acosh addmargins aggregate agrep "
             "alarm alias alist all anova any aov aperm append apply approx "
             "approxfun apropos ar args arima array arrows asin asinh assign "
             "assocplot atan atanh attach attr attributes autoload autoloader "
             "ave axis backsolve barplot basename beta bindtextdomain binomial "
             "biplot bitmap bmp body box boxplot bquote break browser builtins "
             "bxp by bzfile c call cancor capabilities casefold cat category "
             "cbind ccf ceiling character charmatch chartr chol choose chull "
             "citation class close cm cmdscale codes coef coefficients col "
             "colnames colors colorspaces colours comment complex confint "
             "conflicts contour contrasts contributors convolve cophenetic "
             "coplot cor cos cosh cov covratio cpgram crossprod cummax cummin "
             "cumprod cumsum curve cut cutree cycle data dataentry date dbeta "
             "dbinom dcauchy dchisq de debug debugger decompose delay deltat "
             "demo dendrapply density deparse deriv det detach determinant "
             "deviance dexp df dfbeta dfbetas dffits dgamma dgeom dget dhyper "
             "diag diff diffinv difftime digamma dim dimnames dir dirname "
             "dist dlnorm dlogis dmultinom dnbinom dnorm dotchart double "
             "dpois dput drop dsignrank dt dump dunif duplicated dweibull "
             "dwilcox eapply ecdf edit effects eigen emacs embed end "
             "environment eval evalq example exists exp expression factanal "
             "factor factorial family fft fifo file filter find fitted fivenum "
             "fix floor flush for force formals format formula forwardsolve "
             "fourfoldplot frame frequency ftable function gamma gaussian gc "
             "gcinfo gctorture get getenv geterrmessage gettext gettextf getwd "
             "gl glm globalenv gray grep grey grid gsub gzcon gzfile hat "
             "hatvalues hcl hclust head heatmap help hist history hsv "
             "httpclient iconv iconvlist identical identify if ifelse image "
             "influence inherits integer integrate interaction interactive "
             "intersect invisible isoreg jitter jpeg julian kappa kernapply "
             "kernel kmeans knots kronecker ksmooth labels lag lapply layout "
             "lbeta lchoose lcm legend length letters levels lfactorial "
             "lgamma library licence license line lines list lm load "
             "loadhistory loadings local locator loess log logb logical "
             "loglin lowess ls lsfit machine mad mahalanobis makepredictcall "
             "manova mapply match matlines matplot matpoints matrix max mean "
             "median medpolish menu merge message methods mget min missing "
             "mode monthplot months mosaicplot mtext mvfft names napredict "
             "naprint naresid nargs nchar ncol next nextn ngettext nlevels nlm "
             "nls noquote nrow numeric objects offset open optim optimise "
             "optimize options order ordered outer pacf page pairlist pairs "
             "palette par parse paste pbeta pbinom pbirthday pcauchy pchisq "
             "pdf pentagamma person persp pexp pf pgamma pgeom phyper pi pico "
             "pictex pie piechart pipe plclust plnorm plogis plot pmatch pmax "
             "pmin pnbinom png pnorm points poisson poly polygon polym "
             "polyroot postscript power ppoints ppois ppr prcomp predict "
             "preplot pretty princomp print prmatrix prod profile profiler "
             "proj promax prompt provide psigamma psignrank pt ptukey punif "
             "pweibull pwilcox q qbeta qbinom qbirthday qcauchy qchisq qexp qf "
             "qgamma qgeom qhyper qlnorm qlogis qnbinom qnorm qpois qqline "
             "qqnorm qqplot qr qsignrank qt qtukey quantile quarters quasi "
             "quasibinomial quasipoisson quit qunif quote qweibull qwilcox "
             "rainbow range rank raw rbeta rbind rbinom rcauchy rchisq "
             "readline real recover rect reformulate regexpr relevel remove "
             "reorder rep repeat replace replicate replications require "
             "reshape resid residuals restart return rev rexp rf rgamma rgb "
             "rgeom rhyper rle rlnorm rlogis rm rmultinom rnbinom rnorm round "
             "row rownames rowsum rpois rsignrank rstandard rstudent rt rug "
             "runif runmed rweibull rwilcox sample sapply save savehistory "
             "scale scan screen screeplot sd search searchpaths seek segments "
             "seq sequence serialize setdiff setequal setwd shell sign signif "
             "sin single sinh sink smooth solve sort source spectrum spline "
             "splinefun split sprintf sqrt stack stars start stderr stdin "
             "stdout stem step stepfun stl stop stopifnot str strftime "
             "strheight stripchart strptime strsplit strtrim structure "
             "strwidth strwrap sub subset substitute substr substring sum "
             "summary sunflowerplot supsmu svd sweep switch symbols symnum "
             "system t table tabulate tail tan tanh tapply tempdir tempfile "
             "termplot terms tetragamma text time title toeplitz tolower "
             "topenv toupper trace traceback transform trigamma trunc truncate "
             "try ts tsdiag tsp typeof unclass undebug union unique uniroot "
             "unix unlink unlist unname unserialize unsplit unstack untrace "
             "unz update upgrade url var varimax vcov vector version vi "
             "vignette warning warnings weekdays weights which while window "
             "windows with write wsbrowser xedit xemacs xfig xinch xor xtabs "
             "xyinch yinch zapsmall")
R_KEYWORDS3 = ("acme aids aircondit amis aml banking barchart barley beaver "
                "bigcity boot brambles breslow bs bwplot calcium cane "
                "capability cav censboot channing city claridge cloth cloud "
                "coal condense contourplot control corr darwin densityplot "
                "dogs dotplot ducks empinf envelope environmental ethanol fir "
                "frets gpar grav gravity grob hirose histogram islay knn "
                "larrows levelplot llines logit lpoints lsegments lset ltext "
                "lvqinit lvqtest manaus melanoma melanoma motor multiedit "
                "neuro nitrofen nodal ns nuclear oneway parallel paulsen "
                "poisons polar qq qqmath remission rfs saddle salinity shingle "
                "simplex singer somgrid splom stripplot survival tau tmd "
                "tsboot tuna unit urine viewport wireframe wool xyplot")

#---- Syntax Style Specs ----#
if wx.VERSION >= (2, 9, 0, 0, ''):
    SYNTAX_ITEMS = [ (stc.STC_R_BASEKWORD,  'class_style'), #TODO
                     (stc.STC_R_COMMENT,    'comment_style'),
                     (stc.STC_R_DEFAULT,    'default_style'),
                     (stc.STC_R_IDENTIFIER, 'default_style'),
                     (stc.STC_R_INFIX,      'default_style'), #TODO
                     (stc.STC_R_INFIXEOL,   'default_style'), #TODO
                     (stc.STC_R_KWORD,      'keyword_style'),
                     (stc.STC_R_NUMBER,     'number_style'),
                     (stc.STC_R_OPERATOR,   'operator_style'),
                     (stc.STC_R_OTHERKWORD, 'keyword2_style'),
                     (stc.STC_R_STRING,     'string_style'),
                     (stc.STC_R_STRING2,    'char_style')] #TODO
else:
    SYNTAX_ITEMS = [ (STC_S_DEFAULT,   'default_style'),
                     (STC_S_COMMENT,   'comment_style'),
                     (STC_S_NUMBER,    'number_style'),
                     (STC_S_STRING,    'string_style'),
                     (STC_S_STRINGEOL, 'stringeol_style'),
                     (STC_S_OPERATOR,  'operator_style'),
                     (STC_S_KEYWORD,   'keyword_style') ]

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for R and S""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        if wx.VERSION >= (2, 9, 0, 0, ''):
            self.SetLexer(stc.STC_LEX_R)
        else:
            self.SetLexer(stc.STC_LEX_CONTAINER)
            self.RegisterFeature(synglob.FEATURE_STYLETEXT, StyleText)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        if wx.VERSION >= (2, 9, 0, 0, ''):
            return [(0, R_KEYWORDS), (1, R_KEYWORDS2), (2, R_KEYWORDS3)]
        else:
            return [(1, KEYWORDS)]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u"#",]

#-----------------------------------------------------------------------------#

def StyleText(_stc, start, end):
    """Style the text
    @param _stc: Styled text control instance
    @param start: Start position
    @param end: end position
    @todo: performance improvements
    @todo: style errors caused by unicode characters (related to internal utf8)

    """
    cpos = 0
    _stc.StartStyling(cpos, 0x1f)
    lexer = get_lexer_by_name("s")
    is_wineol = _stc.GetEOLMode() == stc.STC_EOL_CRLF
    for token, txt in lexer.get_tokens(_stc.GetTextRange(0, end)):
        style = TOKEN_MAP.get(token, STC_S_DEFAULT)

        tlen = len(txt)

        # Account for \r\n end of line characters
        if is_wineol and "\n" in txt:
            tlen += txt.count("\n")

        if tlen:
            _stc.SetStyling(tlen, style)
        cpos += tlen
        _stc.StartStyling(cpos, 0x1f)

#-----------------------------------------------------------------------------#

TOKEN_MAP = { Token.Literal.String  : STC_S_STRING,
              Token.Comment         : STC_S_COMMENT,
              Token.Comment.Single  : STC_S_COMMENT,
              Token.Operator        : STC_S_OPERATOR,
              Token.Punctuation     : STC_S_OPERATOR,
              Token.Number          : STC_S_NUMBER,
              Token.Literal.Number  : STC_S_NUMBER,
              Token.Keyword         : STC_S_KEYWORD,
              Token.Keyword.Constant: STC_S_KEYWORD }
              


###############################################################################
# Name: django.py                                                             #
# Purpose: Define Django syntax for highlighting and other features           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: django.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Django Templates.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _django.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc
from pygments.token import Token
from pygments.lexers import get_lexer_by_name

#Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#
# Style Id's

STC_DJANGO_DEFAULT, \
STC_DJANGO_COMMENT, \
STC_DJANGO_NUMBER, \
STC_DJANGO_STRING, \
STC_DJANGO_STRINGEOL, \
STC_DJANGO_SCALAR, \
STC_DJANGO_OPERATOR, \
STC_DJANGO_PREPROCESSOR, \
STC_DJANGO_ATTRIBUTE, \
STC_DJANGO_TAG, \
STC_DJANGO_BUILTIN, \
STC_DJANGO_KEYWORD = range(12)

#-----------------------------------------------------------------------------#

# Python Keywords
KEYWORDS = ("true false undefined null in as reversed recursive not and or is "
            "if else import with loop block forloop")

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (STC_DJANGO_DEFAULT, 'default_style'),
                 (STC_DJANGO_COMMENT, 'comment_style'),
                 (STC_DJANGO_NUMBER, 'number_style'),
                 (STC_DJANGO_STRING, 'string_style'),
                 (STC_DJANGO_STRINGEOL, 'stringeol_style'),
                 (STC_DJANGO_SCALAR, 'scalar_style'),
                 (STC_DJANGO_OPERATOR, 'operator_style'),
                 (STC_DJANGO_PREPROCESSOR, 'pre_style'),
                 (STC_DJANGO_ATTRIBUTE, 'keyword2_style'),
                 (STC_DJANGO_TAG,     'keyword_style'),       # Need new tag
                 (STC_DJANGO_BUILTIN, 'keyword4_style'),
                 (STC_DJANGO_KEYWORD, 'keyword_style'), ]

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Django""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_CONTAINER)
        self.RegisterFeature(synglob.FEATURE_STYLETEXT, StyleText)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [(1, KEYWORDS)]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u"#",]

#-----------------------------------------------------------------------------#

def StyleText(stc, start, end):
    """Style the text
    @param stc: Styled text control instance
    @param start: Start position
    @param end: end position

    """
    cpos = 0
    stc.StartStyling(cpos, 0x1f)
    lexer = get_lexer_by_name("html+django")
    doctxt = stc.GetTextRange(0, end)
    wineol = stc.GetEOLChar() == "\r\n"

    # Need to convert to UTF-8 in order to calculate
    # correct positions in STC.
    try:
        doctxt = doctxt.encode('utf-8')
    except:
        pass

    for token, txt in lexer.get_tokens(doctxt):
#        print token, txt
        style = TOKEN_MAP.get(token, STC_DJANGO_DEFAULT)
        if style == STC_DJANGO_PREPROCESSOR and txt.startswith(u'#'):
            style = STC_DJANGO_COMMENT
#        elif style == STC_DJANGO_STRING and txt[-1] not in '"\'':
#            style = STC_DJANGO_STRINGEOL

        tlen = len(txt)
        if wineol and "\n" in txt:
            tlen += txt.count("\n")

        if tlen:
            stc.SetStyling(tlen, style)
        cpos += tlen
        stc.StartStyling(cpos, 0x1f)

#-----------------------------------------------------------------------------#

TOKEN_MAP = { Token.Literal.String : STC_DJANGO_STRING,
              Token.Comment.Preproc : STC_DJANGO_PREPROCESSOR,
              Token.Comment : STC_DJANGO_COMMENT,
              Token.Name.Builtin : STC_DJANGO_BUILTIN,
              Token.Operator : STC_DJANGO_OPERATOR,
              Token.Punctuation : STC_DJANGO_OPERATOR,
              Token.Number : STC_DJANGO_NUMBER,
              Token.Keyword : STC_DJANGO_KEYWORD,
              Token.Name.Attribute : STC_DJANGO_ATTRIBUTE,
              Token.String.Interpol : STC_DJANGO_SCALAR,
              Token.Name.Tag : STC_DJANGO_TAG }
              


###############################################################################
# Name: panelbox.py                                                           #
# Purpose: Advanced listbox control                                           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Editra Control Library: PanelBox

Class PanelBox:

ListBox like container class that accepts custom panels as items in the list.

Class PanelBoxItemBase:

Base class for all PanelBoxItems. Provides the basic functionality for a custom
subclass to interact with the PanelBox container list.

Class PanelBoxItem:

Simple PanelBoxItem that has support for displaying an Icon, Main text, and a
a user defined sub item.

+-------------------------+
|                         |
| ICON   label            |
|        sub item         |
+-------------------------+

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: panelbox.py 67123 2011-03-04 00:02:35Z CJP $"
__revision__ = "$Revision: 67123 $"

#--------------------------------------------------------------------------#
# Imports
import wx
import wx.lib.scrolledpanel as scrolled

#--------------------------------------------------------------------------#

edEVT_ITEM_SELECTED = wx.NewEventType()
EVT_ITEM_SELECTED = wx.PyEventBinder(edEVT_ITEM_SELECTED, 1)

class PanelBoxEventEvent(wx.PyCommandEvent):
    """Panel Box Event Object"""
    pass

#--------------------------------------------------------------------------#

class PanelBox(scrolled.ScrolledPanel):
    """Scrolled container window for managing and displaying PanelBox items"""
    def __init__(self, parent, id=wx.ID_ANY, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.HSCROLL|wx.VSCROLL,
                 name=u"PanelBox"):
        super(PanelBox, self).__init__(parent, id, pos, size, style, name)

        # Attributes
        self._items = list()
        self._last_sel = -1
        self._sizer = wx.BoxSizer(wx.VERTICAL)

        # Setup
        bkgrnd = wx.SystemSettings.GetColour(wx.SYS_COLOUR_LISTBOX)
        self.SetBackgroundColour(bkgrnd)

        self.SetSizer(self._sizer)
        self.SetAutoLayout(True)

        self.SetupScrolling()

        # Event Handlers
#        self.Bind(wx.EVT_KEY_UP, self.OnNavigate)
        self.GetParent().Bind(wx.EVT_KEY_UP, self.OnNavigate)

    #---- Event Handlers ----#

    def OnItemClicked(self, evt):
        """Callback from when children are clicked on
        @param evt: wx.MouseEvent

        """
        item = evt.GetEventObject()
        selected = item.IsSelected()
        idx = self.FindIndex(item)
        if idx == -1:
            return

        if evt.CmdDown():
            # Add/Remove from selection
            item.SetSelection(not selected)
        elif evt.ShiftDown():
            # Select all items between this item and the next selected one
            if idx < self._last_sel:
                inc = -1
            else:
                inc = 1

            for index in range(self._last_sel, idx + inc, inc):
                self.SetSelection(index, True)
        else:
            # Move selection to this item
            self.ClearSelections()

            if not selected:
                item.SetSelection(True)

        if not selected:
            self._last_sel = idx
        else:
            self._last_sel = -1

    def OnNavigate(self, evt):
        """Handle navigation key events"""
        key_code = evt.GetKeyCode()
        nsel = None
        isup = False
        if key_code == wx.WXK_UP:
            if self._last_sel in (0, -1):
                nsel = len(self._items) - 1
            else:
                nsel = self._last_sel - 1
            isup = True
        elif key_code == wx.WXK_DOWN:
            if self._last_sel in (-1, len(self._items) - 1):
                nsel = 0
            else:
                nsel = self._last_sel + 1
        else:
            evt.Skip()
            return

        if evt.ShiftDown():
            self.SetSelection(nsel, True)
        else:
            self.ClearSelections()
            self.SetSelection(nsel, True)

        evt.Skip()

    #---- Public Api ----#
    def AppendItem(self, item):
        """Append an item to the list
        @param item: PanelBoxItem

        """
        self._items.append(item)
        self._sizer.Add(item, 0, wx.EXPAND)
        item.Realize()

    def ClearSelections(self):
        """Unselect all items"""
        for item in self._items:
            item.SetSelection(False)

    def DeleteAllItems(self):
        """Delete all the items in the list"""
        for item in self._items:
            self._sizer.Remove(item)
            try:
                item.Destroy()
            except wx.PyDeadObjectError:
                pass

        del self._items
        self._items = list()
        self.Layout()

    def FindIndex(self, item):
        """Find the index of a given L{PanelBoxItem}
        @param item: instance of PanelBoxItemBase
        @return: int (-1 on failure)

        """
        for idx, pbitem in enumerate(self._items):
            if pbitem is item:
                return idx
        else:
            return -1

    def GetItemCount(self):
        """Get the number of items in the control
        @return: int

        """
        return len(self._items)

    def GetItems(self):
        """Get the list of items held by this control
        @return: list of PanelBoxItems
        @todo: should probably return a list of shadow items so that orignals
               are not modified.

        """
        return self._items

    def GetSelection(self):
        """Get the (first) selected item"""
        for item in self._items:
            if item.IsSelected():
                return item
        else:
            return None

    def GetSelections(self):
        """Get the list of selected items
        @return: list

        """
        return [item for item in self._items if item.IsSelected()]

    def InsertItem(self, index, item):
        """Insert an item into the list
        @param index: index to insert at
        @param item: PanelBoxItem

        """
        if index <= len(self._items):
            self._items.insert(index, item)
            self._sizer.Insert(index, item, 0, wx.EXPAND)
        else:
            raise IndexError, "Index %d: out of range" % index

    def Remove(self, index):
        """Remove an item from the list
        @param index: item index

        """
        if index < len(self._items):
            item = self._items.pop(index)
            self._sizer.Remove(item)
            self.Layout()
        else:
            raise IndexError, "Index %d: out of range" % index

    def RemoveAll(self):
        """Remove all items from the list"""
        for item in self._items:
            self._sizer.Remove(item)

        del self._items
        self._items = list()

        self.Layout()

    def SetSelection(self, idx, select=True):
        """Set the selection on a given index
        @param idx: int
        @keyword select: bool

        """
        if idx < len(self._items):
            item = self._items[idx]
            item.SetSelection(select)
            self._last_sel = idx
        else:
            raise IndexError, "Index out of range: %d > %d" (idx, len(self._items))

#--------------------------------------------------------------------------#

class PanelBoxItemBase(wx.PyPanel):
    """Base L{PanelBox} Item"""
    def __init__(self, parent):
        """Create a PanelBoxItem"""
        super(PanelBoxItemBase, self).__init__(parent,
                                               style=wx.NO_BORDER|wx.TAB_TRAVERSAL)

        # Attributes
        self._selected = False

        # Event Handlers
        self.Bind(wx.EVT_PAINT, self.OnPaint)
#        self.Bind(wx.EVT_KEY_UP, self.OnKeyUp)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseWheel)

    def _UpdateForeground(self, color):
        """Update foreground colors when selection changes
        @param color: selection color
        @todo: should cache text's original color to restore
               on de-selection.

        """
        if sum(color.Get()[:3]) < (127 * 3):
            ncolor = wx.WHITE
        else:
            ncolor = wx.BLACK

        for child in self.GetChildren():
            if hasattr(child, 'SetForegroundColour') and \
               not isinstance(child, wx.Button):
                child.SetForegroundColour(ncolor)

    def OnKeyUp(self, evt):
        """Handle key navigation events"""
        self.GetParent().OnNavigate(evt)
        evt.Skip()

    def OnLeftUp(self, evt):
        """Handle when the item is clicked on"""
        e_obj = evt.GetEventObject()
        evt.SetEventObject(self)
        self.GetParent().OnItemClicked(evt)
        evt.SetEventObject(e_obj)
        evt.Skip()

    def OnMouseWheel(self, evt):
        """Relay the mouse wheel events to the panel box"""
        self.GetParent().GetEventHandler().ProcessEvent(evt)
        evt.Skip()

    def OnPaint(self, evt):
        """Paint the items background"""
        dc = wx.PaintDC(self)
        rect = self.GetClientRect()

#        if self.IsSelected():
#            color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHT)
#            dc.SetBrush(wx.Brush(color))
#            dc.SetPen(wx.TRANSPARENT_PEN)
#            dc.DrawRectangle(*rect)
#        else:
#            col2 = wx.SystemSettings_GetColour(wx.SYS_COLOUR_LISTBOX)
#            dc.SetBackground(wx.Brush(col2))
#            dc.SetBrush(wx.Brush(col2))
#            dc.SetPen(wx.TRANSPARENT_PEN)
#            dc.DrawRectangle(rect.x, rect.y, rect.width, rect.height - 1)

        pcolor = wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE)
        dc.SetPen(wx.Pen(pcolor))
        dc.DrawLine(rect.x, rect.bottom, rect.right, rect.bottom)

    def IsSelected(self):
        """Is this item selected
        @return: bool

        """
        return self._selected

    def Realize(self):
        """Finalize initialization of the panel item"""
        for child in self.GetChildren():
            child.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)

    def SetSelection(self, select=False):
        """Set the selection state on this item
        @keyword select: bool

        """
        self._selected = select
        if self._selected:
            color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        else:
            color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_LISTBOX)
        self.SetBackgroundColour(color)
        self._UpdateForeground(color)
        self.Refresh()

#--------------------------------------------------------------------------#

class PanelBoxItem(PanelBoxItemBase):
    """L{PanelBox} Item that has an icon, main label text and sub label
    +-------------------------+
    |                         |
    | ICON   label            |
    |        sub item         |
    +-------------------------+

    """
    def __init__(self, parent, bmp=None, label=u'', sub=None):
        """Create teh PanelBoxItem
        @param parent: L{PanelBox}
        @keyword bmp: wx.Bitmap
        @keyword label: string
        @keyword sub: Window object or None

        """
        super(PanelBoxItem, self).__init__(parent)

        # Attributes
        self._bmp = bmp
        self._label = label
        self._sub = sub

        # Layout
        self.__DoLayout()
        self.SetAutoLayout(True)

    def __DoLayout(self):
        """Layout the control"""
        vsizer = wx.BoxSizer(wx.VERTICAL)
        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        hsizer.Add((8, 8), 0)
        if self._bmp is not None:
            self._bmp = wx.StaticBitmap(self, bitmap=self._bmp)
            hsizer.Add(self._bmp, 0, wx.ALIGN_CENTER_VERTICAL)
            hsizer.Add((5, 5), 0)

        # Add Label Text
        isizer = wx.BoxSizer(wx.VERTICAL)
        self._label = wx.StaticText(self, label=self._label)
        isizer.Add(self._label, 0, wx.ALIGN_LEFT)
        if self._sub is not None:
            isizer.Add((3, 3), 0)

        # Add Subwindow if one is defined
        if self._sub is not None:
            self._sub.Reparent(self)
            s_sizer = wx.BoxSizer(wx.HORIZONTAL)
            s_sizer.Add(self._sub, 1, wx.EXPAND)
            isizer.Add(s_sizer, 1, wx.EXPAND)

        hsizer.Add(isizer, 1, wx.ALIGN_CENTER_VERTICAL)
        hsizer.Add((8, 8), 0)
        vsizer.AddMany([((8, 8), 0), (hsizer, 0, wx.EXPAND), ((8, 8), 0)])
        self.SetSizer(vsizer)

    def SetBitmap(self, bmp):
        """Set the items image
        param bmp: wx.Bitmap

        """
        self._bmp.SetBitmap(bmp)
        self._bmp.Refresh()
        self.Layout()

    def SetLabel(self, lbl):
        """Set the label text
        @param lbl: string

        """
        self._lbl.SetLabel(lbl)
        self._lbl.Refresh()
        self.Layout()

    def SetSecondaryCtrl(self, ctrl):
        """Set the secondary control
        @param ctrl: wxWindow

        """
        pass



###############################################################################
# Name: ProgressStatusBarDemo.py                                              #
# Purpose: Test and demo of ProgressStatusBar class.                          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Test file for testing the ProgressStatusBar control

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ProgressStatusBarDemo.py 61806 2009-09-02 01:31:01Z CJP $"
__revision__ = "$Revision: 61806 $"

#-----------------------------------------------------------------------------#
# Imports
import os
import sys
import wx

# Put local package on the path
#sys.path.insert(0, os.path.abspath('../../src'))
import eclib

#-----------------------------------------------------------------------------#

class TestPanel(wx.Panel):
    def __init__(self, parent, log):
        wx.Panel.__init__(self, parent, wx.ID_ANY, size=(500,500))

        # Attributes
        self.log = log
        self.button = wx.Button(self, label="Show ProgressStatusBar Demo")

        # Layout
        self.__DoLayout()

        # Event Handlers
        self.button.Bind(wx.EVT_BUTTON, self.OnButton)

    def __DoLayout(self):
        """Layout the panel"""
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.AddStretchSpacer()
        sizer.Add(self.button, 0, wx.ALIGN_CENTER)
        sizer.AddStretchSpacer()

        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        hsizer.AddStretchSpacer()
        hsizer.Add(sizer, 0, wx.ALIGN_CENTER)
        hsizer.AddStretchSpacer()

        self.SetSizer(hsizer)

    def OnButton(self, evt):
        frame = PBSDemoFrame()
        frame.Show()

#----------------------------------------------------------------------

ID_START = wx.NewId()
ID_START_BUSY = wx.NewId()

class PBSDemoFrame(wx.Frame):
    def __init__(self):
        wx.Frame.__init__(self, None, title="ProgressStatusBar Demo")

        # Attributes
        self.panel = PBSDemoPanel(self)
        self.timer = wx.Timer(self)

        # Layout
        self.__DoLayout()

        # Event Handlers
        self.Bind(wx.EVT_BUTTON, self.OnButton)
        self.Bind(wx.EVT_TIMER, self.OnTimer)

    def __del__(self):
        if self.timer.IsRunning():
            self.timer.Stop()

    def __DoLayout(self):
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.panel, 1, wx.EXPAND)
        self.SetSizer(sizer)
        # NOTE: must use SB_FLAT style
        # TODO: fix this in control to enforce this req.
        statb = eclib.ProgressStatusBar(self, style=wx.SB_FLAT)
        statb.SetRange(1024)
        self.SetStatusBar(statb)

    def OnButton(self, evt):
        e_id = evt.GetId()
        statb = self.GetStatusBar()
        self.panel.SetButtonState(e_id)

        if e_id == wx.ID_STOP:
            statb.SetStatusText(u"", 0)
            if self.timer.IsRunning():
                self.timer.Stop()
            statb.Stop()
            
        elif e_id == ID_START:
            # Start timer to simulate updates with
            statb.ShowProgress()
            self.timer.Start(125)
        elif e_id == ID_START_BUSY:
            statb.StartBusy()
            self.SetStatusText("Busy...", 0)
        else:
            evt.Skip()

    def OnTimer(self, evt):
        statb = self.GetStatusBar()
        cprog = statb.GetProgress()
        cprog += 8
        if cprog > 1024:
            cprog = 0
        statb.SetProgress(cprog)
        self.SetStatusText("Processing %d of 1024" % cprog, 0)

#----------------------------------------------------------------------

class PBSDemoPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent)

        # Attributes
        self.startb = wx.Button(self, ID_START, label="Start")
        self.startb.SetToolTipString("Click to start progress bar")
        self.busyb = wx.Button(self, ID_START_BUSY, label="Start Busy")
        self.busyb.SetToolTipString("Click to start progress bar in 'Busy' mode")
        self.stopb = wx.Button(self, wx.ID_STOP, label="Stop")
        self.stopb.SetToolTipString("Stop and hide the progress gauge")
        self.stopb.Disable()

        # Layout
        self.__DoLayout()

    def __DoLayout(self):
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.AddMany([((-1, 5), 0, wx.EXPAND),
                       (self.startb, 0, wx.ALIGN_CENTER_VERTICAL),
                       (self.busyb, 0, wx.ALIGN_CENTER_VERTICAL),
                       (self.stopb, 0, wx.ALIGN_CENTER_VERTICAL),
                       ((-1, 5), 0, wx.EXPAND)])
        vsizer = wx.BoxSizer(wx.VERTICAL)
        vsizer.Add(sizer, 1, wx.ALIGN_CENTER)
        self.SetSizer(vsizer)

    def SetButtonState(self, state):
        if state in (ID_START, ID_START_BUSY):
            self.busyb.Disable()
            self.startb.Disable()
            self.stopb.Enable()
        elif state == wx.ID_STOP:
            self.busyb.Enable()
            self.startb.Enable()
            self.stopb.Disable()
        else:
            pass

#----------------------------------------------------------------------

def runTest(frame, nb, log):
    win = TestPanel(nb, log)
    return win

class TestLog:
    def __init__(self):
        pass

    def write(self, msg):
        print msg

#----------------------------------------------------------------------

overview = eclib.pstatbar.__doc__
title = "ProgressStatusBar"

#-----------------------------------------------------------------------------#
if __name__ == '__main__':
    try:
        import run
    except ImportError:
        app = wx.PySimpleApp(False)
        frame = wx.Frame(None, title="ProgressStatusBar Demo")
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(TestPanel(frame, TestLog()), 1, wx.EXPAND)
        frame.CreateStatusBar()
        frame.SetSizer(sizer)
        frame.Show()
        app.MainLoop()
    else:
        run.main(['', os.path.basename(sys.argv[0])] + sys.argv[1:])



import copy

class Error(Exception):
    """Base exception class for the dexml module."""
    pass

class ParseError(Error):
    """Exception raised when XML could not be parsed into objects."""
    pass

class RenderError(Error):
    """Exception raised when object could not be rendered into XML."""
    pass

class XmlError(Error):
    """Exception raised to encapsulate errors from underlying XML parser."""
    pass

class PARSE_DONE:
    """Constant returned by a Field when it has finished parsing."""
    pass
class PARSE_MORE:
    """Constant returned by a Field when it wants additional nodes to parse."""
    pass
class PARSE_SKIP:
    """Constant returned by a Field when it cannot parse the given node."""
    pass
class PARSE_CHILDREN:
    """Constant returned by a Field to parse children from its container tag."""
    pass

class Meta:
    """Class holding meta-information about a dexml.Model subclass.

    Each dexml.Model subclass has an attribute 'meta' which is an instance
    of this class.  That instance holds information about how to model 
    corresponds to XML, such as its tagname, namespace, and error handling
    semantics.  You would not ordinarily create an instance of this class;
    instead let the ModelMetaclass create one automatically.

    These attributes control how the model corresponds to the XML:

        * tagname:  the name of the tag representing this model
        * namespace:  the XML namespace in which this model lives

    These attributes control parsing/rendering behaviour:

        * namespace_prefix:  the prefix to use for rendering namespaced tags
        * ignore_unknown_elements:  ignore unknown elements when parsing
        * case_sensitive:    match tag/attr names case-sensitively
        * order_sensitive:   match child tags in order of field definition

    """

    _defaults = {"tagname":None,
                 "namespace":None,
                 "namespace_prefix":None,
                 "ignore_unknown_elements":True,
                 "case_sensitive":True,
                 "order_sensitive":True}

    def __init__(self,name,meta_attrs):
        for (attr,default) in self._defaults.items():
            setattr(self,attr,meta_attrs.get(attr,default))
        if self.tagname is None:
            self.tagname = name

def _meta_attributes(meta):
    """Extract attributes from a "meta" object."""
    meta_attrs = {}
    if meta:
        for attr in dir(meta):
            if not attr.startswith("_"):
                meta_attrs[attr] = getattr(meta,attr)
    return meta_attrs

class ModelMetaclass(type):
    """Metaclass for dexml.Model and subclasses.

    This metaclass is responsible for introspecting Model class definitions
    and setting up appropriate default behaviours.  For example, this metaclass
    sets a Model's default tagname to be equal to the declared class name.
    """

    instances = {}

    def __new__(mcls,name,bases,attrs):
        cls = super(ModelMetaclass,mcls).__new__(mcls,name,bases,attrs)
        #  Don't do anything if it's not a subclass of Model
        parents = [b for b in bases if isinstance(b, ModelMetaclass)]
        # HACK
        import fields

        if not parents:
            return cls
        #  Set up the cls.meta object, inheriting from base classes
        meta_attrs = {}
        for base in bases:
            if isinstance(base,ModelMetaclass) and hasattr(base,"meta"):
                meta_attrs.update(_meta_attributes(base.meta))
        meta_attrs.pop("tagname",None)
        meta_attrs.update(_meta_attributes(attrs.get("meta",None)))
        cls.meta = Meta(name,meta_attrs)
        #  Create ordered list of field objects, telling each about their
        #  name and containing class.  Inherit fields from base classes
        #  only if not overridden on the class itself.
        base_fields = {}
        for base in bases:
            if not isinstance(base,ModelMetaclass):
                continue
            for field in base._fields:
                if field.field_name not in base_fields:
                    field = copy.copy(field)
                    field.model_class = cls
                    base_fields[field.field_name] = field
        cls_fields = []
        for (name,value) in attrs.iteritems():
            if isinstance(value,fields.Field):
                base_fields.pop(name,None)
                value.field_name = name
                value.model_class = cls
                cls_fields.append(value)
        cls._fields = base_fields.values() + cls_fields
        cls._fields.sort(key=lambda f: f._order_counter)
        #  Register the new class so we can find it by name later on
        mcls.instances[(cls.meta.namespace,cls.meta.tagname)] = cls
        return cls

    @classmethod
    def find_class(mcls,tagname,namespace=None):
        """Find dexml.Model subclass for the given tagname and namespace."""
        return mcls.instances.get((namespace,tagname))



###############################################################################
# Name: apache.py                                                             #
# Purpose: Define Apache syntax for highlighting and other features           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: apache.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Apache Configuration Files

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _apache.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
DIRECTIVES = (0, 'acceptmutex acceptpathinfo accessconfig accessfilename '
                 'action addalt addaltbyencoding addaltbytype addcharset '
                 'adddefaultcharset adddescription addencoding addhandler '
                 'addicon addiconbyencoding addiconbytype addinputfilter '
                 'addlanguage addmodule addmoduleinfo addoutputfilter '
                 'addoutputfilterbytype addtype agentlog alias aliasmatch all '
                 'allow allowconnect allowencodedslashes allowoverride '
                 'anonymous anonymous_authoritative anonymous_logemail '
                 'anonymous_mustgiveemail anonymous_nouserid '
                 'anonymous_verifyemail assignuserid authauthoritative '
                 'authdbauthoritative authdbgroupfile authdbmauthoritative '
                 'authdbmgroupfile authdbmtype authdbmuserfile authdbuserfile '
                 'authdigestalgorithm authdigestdomain authdigestfile '
                 'authdigestgroupfile authdigestnccheck authdigestnonceformat '
                 'authdigestnoncelifetime authdigestqop authdigestshmemsize '
                 'authgroupfile authldapauthoritative authldapbinddn '
                 'authldapbindpassword authldapcharsetconfig '
                 'authldapcomparednonserver authldapdereferencealiases '
                 'authldapenabled authldapfrontpagehack authldapgroupattribute '
                 'authldapgroupattributeisdn authldapremoteuserisdn '
                 'authldapurl authname authtype authuserfile bindaddress '
                 'browsermatch browsermatchnocase bs2000account bufferedlogs '
                 'cachedefaultexpire cachedirlength cachedirlevels '
                 'cachedisable cacheenable cacheexpirycheck cachefile '
                 'cacheforcecompletion cachegcclean cachegcdaily '
                 'cachegcinterval cachegcmemusage cachegcunused '
                 'cacheignorecachecontrol cacheignoreheaders '
                 'cacheignorenolastmod cachelastmodifiedfactor cachemaxexpire '
                 'cachemaxfilesize cacheminfilesize cachenegotiateddocs '
                 'cacheroot cachesize cachetimemargin cgimapextension '
                 'charsetdefault charsetoptions charsetsourceenc checkspelling '
                 'childperuserid clearmodulelist contentdigest cookiedomain '
                 'cookieexpires cookielog cookiename cookiestyle '
                 'cookietracking coredumpdirectory customlog dav '
                 'davdepthinfinity davlockdb davmintimeout defaulticon '
                 'defaultlanguage defaulttype define deflatebuffersize '
                 'deflatecompressionlevel deflatefilternote deflatememlevel '
                 'deflatewindowsize deny directory directoryindex '
                 'directorymatch directoryslash documentroot dumpioinput '
                 'dumpiooutput enableexceptionhook enablemmap enablesendfile '
                 'errordocument errorlog example expiresactive expiresbytype '
                 'expiresdefault extendedstatus extfilterdefine '
                 'extfilteroptions fancyindexing fileetag files filesmatch '
                 'forcelanguagepriority forcetype forensiclog from group '
                 'header headername hostnamelookups identitycheck ifdefine '
                 'ifmodule imapbase imapdefault imapmenu include indexignore '
                 'indexoptions indexorderdefault isapiappendlogtoerrors '
                 'isapiappendlogtoquery isapicachefile isapifakeasync '
                 'isapilognotsupported isapireadaheadbuffer keepalive '
                 'keepalivetimeout languagepriority ldapcacheentries '
                 'ldapcachettl ldapconnectiontimeout ldapopcacheentries '
                 'ldapopcachettl ldapsharedcachefile ldapsharedcachesize '
                 'ldaptrustedca ldaptrustedcatype limit limitexcept '
                 'limitinternalrecursion limitrequestbody limitrequestfields '
                 'limitrequestfieldsize limitrequestline limitxmlrequestbody '
                 'listen listenbacklog loadfile loadmodule location '
                 'locationmatch lockfile logformat loglevel maxclients '
                 'maxkeepaliverequests maxmemfree maxrequestsperchild '
                 'maxrequestsperthread maxspareservers maxsparethreads '
                 'maxthreads maxthreadsperchild mcachemaxobjectcount '
                 'mcachemaxobjectsize mcachemaxstreamingbuffer '
                 'mcacheminobjectsize mcacheremovalalgorithm mcachesize '
                 'metadir metafiles metasuffix mimemagicfile minspareservers '
                 'minsparethreads mmapfile modmimeusepathinfo multiviewsmatch '
                 'namevirtualhost nocache noproxy numservers nwssltrustedcerts '
                 'nwsslupgradeable options order passenv pidfile port '
                 'protocolecho proxy proxybadheader proxyblock proxydomain '
                 'proxyerroroverride proxyiobuffersize proxymatch '
                 'proxymaxforwards proxypass proxypassreverse '
                 'proxypreservehost proxyreceivebuffersize proxyremote '
                 'proxyremotematch proxyrequests proxytimeout proxyvia qsc '
                 'readmename redirect redirectmatch redirectpermanent '
                 'redirecttemp refererignore refererlog removecharset '
                 'removeencoding removehandler removeinputfilter '
                 'removelanguage removeoutputfilter removetype requestheader '
                 'require resourceconfig rewritebase rewritecond rewriteengine '
                 'rewritelock rewritelog rewriteloglevel rewritemap '
                 'rewriteoptions rewriterule rlimitcpu rlimitmem rlimitnproc '
                 'satisfy scoreboardfile script scriptalias scriptaliasmatch '
                 'scriptinterpretersource scriptlog scriptlogbuffer '
                 'scriptloglength scriptsock securelisten sendbuffersize '
                 'serveradmin serveralias serverlimit servername serverpath '
                 'serverroot serversignature servertokens servertype setenv '
                 'setenvif setenvifnocase sethandler setinputfilter '
                 'setoutputfilter singlelisten ssiendtag ssierrormsg '
                 'ssistarttag ssitimeformat ssiundefinedecho '
                 'sslcacertificatefile sslcacertificatepath '
                 'sslcarevocationfile sslcarevocationpath '
                 'sslcertificatechainfile sslcertificatefile '
                 'sslcertificatekeyfile sslciphersuite sslengine sslmutex '
                 'ssloptions sslpassphrasedialog sslprotocol '
                 'sslproxycacertificatefile sslproxycacertificatepath '
                 'sslproxycarevocationfile sslproxycarevocationpath '
                 'sslproxyciphersuite sslproxyengine '
                 'sslproxymachinecertificatefile '
                 'sslproxymachinecertificatepath sslproxyprotocol '
                 'sslproxyverify sslproxyverifydepth sslrandomseed sslrequire '
                 'sslrequiressl sslsessioncache sslsessioncachetimeout '
                 'sslusername sslverifyclient sslverifydepth startservers '
                 'startthreads suexecusergroup threadlimit threadsperchild '
                 'threadstacksize timeout transferlog typesconfig unsetenv '
                 'usecanonicalname user userdir virtualdocumentroot '
                 'virtualdocumentrootip virtualhost virtualscriptalias '
                 'virtualscriptaliasip win32disableacceptex xbithack')

PARAMS = (1, 'on off standalone inetd force-response-1.0 downgrade-1.0 '
             'nokeepalive indexes includes followsymlinks none x-compress '
             'x-gzip warn')

#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [(stc.STC_CONF_COMMENT,  'comment_style'),
                (stc.STC_CONF_DEFAULT,  'default_style'),
                (stc.STC_CONF_DIRECTIVE, 'keyword_style'),
                (stc.STC_CONF_EXTENSION, 'pre_style'),
                (stc.STC_CONF_IDENTIFIER, 'number_style'),
                (stc.STC_CONF_IP,       'number2_style'),
                (stc.STC_CONF_NUMBER,   'number_style'),
                (stc.STC_CONF_OPERATOR, 'operator_style'),
                (stc.STC_CONF_PARAMETER, 'global_style'),
                (stc.STC_CONF_STRING,   'string_style')]

#---- Extra Properties ----#

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Apache Conf files""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_CONF)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [DIRECTIVES, PARAMS]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'#']



###############################################################################
# Name: ed_statbar.py                                                         #
# Purpose: Custom statusbar with builtin progress indicator                   #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Custom StatusBar for Editra that contains a progress bar that responds to
messages from ed_msg to display progress of different actions.

@summary: Editra's StatusBar class

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ed_statbar.py 70229 2012-01-01 01:27:10Z CJP $"
__revision__ = "$Revision: 70229 $"

#--------------------------------------------------------------------------#
# Imports
import wx
import wx.stc

# Editra Libraries
import ed_glob
import util
import ed_msg
import ed_menu
from syntax.synglob import GetDescriptionFromId
from eclib import ProgressStatusBar, EncodingDialog
from extern.decorlib import anythread

#--------------------------------------------------------------------------#
 
_ = wx.GetTranslation

#--------------------------------------------------------------------------#

class EdStatBar(ProgressStatusBar):
    """Custom status bar that handles dynamic field width adjustment and
    automatic expiration of status messages.

    """
    ID_CLEANUP_TIMER = wx.NewId()
    def __init__(self, parent):
        super(EdStatBar, self).__init__(parent, style=wx.ST_SIZEGRIP)

        # Attributes
        self._pid = parent.GetId() # Save parents id for filtering msgs
        self._widths = list()
        self._cleanup_timer = wx.Timer(self, EdStatBar.ID_CLEANUP_TIMER)
        self._eolmenu = wx.Menu()
        self._lexmenu = None
        self._log = wx.GetApp().GetLog()

        # Setup
        self.SetFieldsCount(6) # Info, vi stuff, line/progress
        self.SetStatusWidths([-1, 90, 40, 40, 40, 155])
        self._eolmenu.Append(ed_glob.ID_EOL_MAC, u"CR",
                             _("Change line endings to %s") % u"CR",
                             kind=wx.ITEM_CHECK)
        self._eolmenu.Append(ed_glob.ID_EOL_WIN, u"CRLF",
                             _("Change line endings to %s") % u"CRLF",
                             kind=wx.ITEM_CHECK)
        self._eolmenu.Append(ed_glob.ID_EOL_UNIX, u"LF",
                             _("Change line endings to %s") % u"LF",
                             kind=wx.ITEM_CHECK)

        # Event Handlers
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy, self)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnLeftDClick)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self.Bind(wx.EVT_TIMER, self.OnExpireMessage,
                  id=EdStatBar.ID_CLEANUP_TIMER)

        # Messages
        ed_msg.Subscribe(self.OnProgress, ed_msg.EDMSG_PROGRESS_SHOW)
        ed_msg.Subscribe(self.OnProgress, ed_msg.EDMSG_PROGRESS_STATE)
        ed_msg.Subscribe(self.OnUpdateText, ed_msg.EDMSG_UI_SB_TXT)
        ed_msg.Subscribe(self.OnUpdateDoc, ed_msg.EDMSG_UI_NB_CHANGED)
        ed_msg.Subscribe(self.OnUpdateDoc, ed_msg.EDMSG_FILE_SAVED)
        ed_msg.Subscribe(self.OnUpdateDoc, ed_msg.EDMSG_FILE_OPENED)
        ed_msg.Subscribe(self.OnUpdateDoc, ed_msg.EDMSG_UI_STC_LEXER)

    def OnDestroy(self, evt):
        """Unsubscribe from messages"""
        if self._lexmenu:
            self._lexmenu.Destroy()
        if self._eolmenu:
            self._eolmenu.Destroy()
        if evt.GetId() == self.GetId():
            ed_msg.Unsubscribe(self.OnProgress)
            ed_msg.Unsubscribe(self.OnUpdateText)
            ed_msg.Unsubscribe(self.OnUpdateDoc)
        evt.Skip()

    def __SetStatusText(self, txt, field):
        """Safe method to use for setting status text with CallAfter.
        @param txt: string
        @param field: int

        """
        try:
            super(EdStatBar, self).SetStatusText(txt, field)
            self.AdjustFieldWidths()

            if field == ed_glob.SB_INFO and txt != u'':
                # Start the expiration countdown
                if self._cleanup_timer.IsRunning():
                    self._cleanup_timer.Stop()
                self._cleanup_timer.Start(10000, True)
        except wx.PyDeadObjectError, wx.PyAssertionError:
            # Getting some odd assertion errors on wxMac so just trap
            # and ignore them for now
            # glyphCount == (text.length()+1)" failed at graphics.cpp(2048)
            # in GetPartialTextExtents()
            pass
        except TypeError, err:
            self._log("[edstatbar][err] Bad status message: %s" % str(txt))
            self._log("[edstatbar][err] %s" % err)

    def AdjustFieldWidths(self):
        """Adjust each field width of status bar basing on the field text
        @return: None

        """
        widths = [-1]
        # Calculate required widths
        # NOTE: Order of fields is important
        for field in [ed_glob.SB_BUFF,
                      ed_glob.SB_LEXER,
                      ed_glob.SB_ENCODING,
                      ed_glob.SB_EOL,
                      ed_glob.SB_ROWCOL]:
            width = self.GetTextExtent(self.GetStatusText(field))[0] + 20
            if width == 20:
                width = 0
            widths.append(width)

        # Adjust widths
        if widths[-1] < 155:
            widths[-1] = 155

        # Only update if there are changes
        if widths != self._widths:
            self._widths = widths
            self.SetStatusWidths(self._widths)

    def GetMainWindow(self):
        """Method required for L{ed_msg.mwcontext}"""
        return self.TopLevelParent

    def OnExpireMessage(self, evt):
        """Handle Expiring the status message when the oneshot timer
        tells us it has expired.

        """
        if evt.GetId() == EdStatBar.ID_CLEANUP_TIMER:
            wx.CallAfter(self.__SetStatusText, u'', ed_glob.SB_INFO)
        else:
            evt.Skip()

    def OnLeftDClick(self, evt):
        """Handlers mouse left double click on status bar
        @param evt: wx.MouseEvent
        @note: Assumes parent is MainWindow instance

        """
        pt = evt.GetPosition()
        if self.GetFieldRect(ed_glob.SB_ROWCOL).Contains(pt):
            mw = self.GetParent()
            mpane = mw.GetEditPane()
            mpane.ShowCommandControl(ed_glob.ID_GOTO_LINE)
        else:
            evt.Skip()

    def OnLeftUp(self, evt):
        """Handle left clicks on the status bar
        @param evt: wx.MouseEvent

        """
        pt = evt.GetPosition()
        if self.GetFieldRect(ed_glob.SB_EOL).Contains(pt):
            rect = self.GetFieldRect(ed_glob.SB_EOL)
            self.PopupMenu(self._eolmenu, (rect.x, rect.y))
        elif self.GetFieldRect(ed_glob.SB_ENCODING).Contains(pt):
            nb = self.GetTopLevelParent().GetNotebook()
            buff = nb.GetCurrentCtrl()
            dlg = EncodingDialog(nb,
                                 msg=_("Change the encoding of the current document."),
                                 title=_("Change Encoding"),
                                 default=buff.GetEncoding())
            bmp = wx.ArtProvider.GetBitmap(str(ed_glob.ID_DOCPROP),
                                           wx.ART_OTHER)
            if bmp.IsOk():
                dlg.SetBitmap(bmp)
            dlg.CenterOnParent()

            # TODO: should add EdFile callbacks for modification events instead
            #       of using explicit statusbar refresh.
            if dlg.ShowModal() == wx.ID_OK:
                buff.SetEncoding(dlg.GetEncoding())
                self.UpdateFields()

            # NOTE: Got an error report about a PyDeadObject error here. The
            #       error does not make any sense since the dialog is not
            #       destroyed or deleted by anything before this. Add validity
            #       check to ensure reference is still valid.
            if dlg:
                dlg.Destroy()
        elif self.GetFieldRect(ed_glob.SB_LEXER).Contains(pt):
            # Change Lexer popup menu
            if self._lexmenu:
                self._lexmenu.Destroy()
            self._lexmenu = wx.Menu()
            ed_menu.EdMenuBar.PopulateLexerMenu(self._lexmenu)
            rect = self.GetFieldRect(ed_glob.SB_LEXER)
            self.PopupMenu(self._lexmenu, (rect.x, rect.y))
        else:
            evt.Skip()

    def OnProgress(self, msg):
        """Set the progress bar's state
        @param msg: Message Object

        """
        mdata = msg.GetData()
        # Don't do anything if the message is not for this frame
        if self._pid != mdata[0]:
            return

        mtype = msg.GetType()
        if mtype == ed_msg.EDMSG_PROGRESS_STATE:
            # May be called from non gui thread so don't do anything with
            # the gui here.
            self.SetProgress(mdata[1])
            self.range = mdata[2]
            if sum(mdata[1:]) == 0:
                self.Stop()
        elif mtype == ed_msg.EDMSG_PROGRESS_SHOW:
            if mdata[1]:
                self.Start(75)
            else:
                # TODO: findout where stray stop event is coming from...
                self.Stop()

    @ed_msg.mwcontext
    def OnUpdateDoc(self, msg):
        """Update document related fields
        @param msg: Message Object

        """
        self.UpdateFields()
        if msg.GetType() == ed_msg.EDMSG_UI_NB_CHANGED:
            wx.CallAfter(self.__SetStatusText, u'', ed_glob.SB_INFO)

    @anythread
    def DoUpdateText(self, msg):
        """Thread safe update of status text. Proxy for OnUpdateText because
        pubsub seems to have issues with passing decorator methods for
        listeners.
        @param msg: Message Object

        """
        # Only process if this status bar is in the active window and shown
        parent = self.GetTopLevelParent()
        if (parent.IsActive() or wx.GetApp().GetTopWindow() == parent):
            field, txt = msg.GetData()
            self.UpdateFields()
            wx.CallAfter(self.__SetStatusText, txt, field)

    def OnUpdateText(self, msg):
        """Update the status bar text based on the received message
        @param msg: Message Object

        """
        self.DoUpdateText(msg)

    def PushStatusText(self, txt, field):
        """Set the status text
        @param txt: Text to put in bar
        @param field: int

        """
        wx.CallAfter(self.__SetStatusText, txt, field)

    def SetStatusText(self, txt, field):
        """Set the status text
        @param txt: Text to put in bar
        @param field: int

        """
        wx.CallAfter(self.__SetStatusText, txt, field)

    def UpdateFields(self):
        """Update document fields based on the currently selected
        document in the editor.
        @postcondition: encoding and lexer fields are updated
        @todo: update when readonly hooks are implemented

        """
        nb = self.GetParent().GetNotebook()
        if nb is None:
            return

        try:
            cbuff = nb.GetCurrentCtrl()
            doc = cbuff.GetDocument()
            wx.CallAfter(self.__SetStatusText, doc.GetEncoding(),
                         ed_glob.SB_ENCODING)
            wx.CallAfter(self.__SetStatusText,
                         GetDescriptionFromId(cbuff.GetLangId()),
                         ed_glob.SB_LEXER)

            eol = { wx.stc.STC_EOL_CR : u"CR",
                    wx.stc.STC_EOL_LF : u"LF",
                    wx.stc.STC_EOL_CRLF : u"CRLF" }
            wx.CallAfter(self.__SetStatusText,
                         eol[cbuff.GetEOLMode()],
                         ed_glob.SB_EOL)

        except wx.PyDeadObjectError:
            # May be called asyncronasly after the control is already dead
            return



###############################################################################
# Name: lout.py                                                               #
# Purpose: Define Lout syntax for highlighting and other features             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: lout.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Lout
@todo: style refinement

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _lout.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
# @ prefixed keywords
LOUT_KW1 = (0, "@OptGall @FontDef @Family @Face @Name @Metrics @ExtraMetrics "
               "@Mapping @Recode @Filter @FilterIn @FilterOut @FilterErr @AL "
               "@Common @Rump @Meld @Insert @OneOf @Next @Plus @Minus @Wide "
               "@High @HShift @VShift @BeginHeaderComponent @Document @TItle "
               "@SetHeaderComponent @ClearHeaderComponent @OneCol @OneRow @Doc "
               "@HMirror @VMirror @HScale @VScale @HCover @VCover @Scale @Text "
               "@KernShrink @HContract @VContract @HLimited @VLimited @HExpand "
               "@VExpand @StartHVSpan @StartHSpan @StartVSpan @HSpan @VSpan "
               "@PAdjust @HAdjust @VAdjust @Rotate @Background @IncludeGraphic "
               "@SysIncludeGraphic @Graphic @LinkSource @LinkDest @URLLink @BI "
               "@PlainGraphic @Verbatim @RawVerbatim @Case @Yield @BackEnd @BL "
               "@Char @Font @Space @YUnit @ZUnit @Break @Underline @SetColour "
               "@SetColor @SetUnderlineColour @SetUnderlineColor @SetTexture "
               "@Outline @Language @CurrLang @CurrFamily @CurrFace @CurrYUnit "
               "@CurrZUnit @LEnv @@A @@B @@C @@D @@E @LClos @@V @LUse @LEO @PP "
               "@Open @Use @NotRevealed @Tagged @Database @SysDatabase @I @B"
               "@Include @SysInclude @IncludeGraphicRepeated @InitialFont "
               "@SysIncludeGraphicRepeated @PrependGraphic @SysPrependGraphic "
               "@Target @Null @PageLabel @Galley @ForceGalley @LInput @Split "
               "@Tag @Key @Optimize @Merge @Enclose @Begin @End @Moment @Tab "
               "@Second @Minute @Hour @Day @Month @Year @Century @WeekDay "
               "@YearDay @DaylightSaving @SetContext @GetContext @Time @List "
               "@EndHeaderComponent @Section @BeginSections @EndNote @Abstract "
               "@AlphaList @Appendix @Author @Figure @Report @OuterNote "
               "@IndentedList @InitialBreak @InitialLanguage InnerNote "
               "@Heading @FootNote @Date @LeftList @LeftNote @ListItem "
               "@RightDisplay @RightNote @EndSections")

# Symbols
LOUT_KW2 = (1, "&&& && & ^// ^/ ^|| ^| ^& // / || |")

# Non @ keywords
LOUT_KW3 = (2, "def langdef force horizontally into extend import export "
               "precedence associativity left right body macro named "
               "compulsory following preceding foll_or_prec now "
               "Base Slope Bold BoldSlope Times Helvetica Courier Palatino "
               "adjust breakstyle clines lines linesep hyphen nonhyphen ragged "
               )

# Document Classes
LOUT_KW4 = (3, "fx vx aformat bformat doc eq graph slides tab text tbl")
#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [(stc.STC_LOUT_COMMENT, 'comment_style'),
                (stc.STC_LOUT_DEFAULT,'default_style'),
                (stc.STC_LOUT_IDENTIFIER, 'default_style'),
                (stc.STC_LOUT_NUMBER, 'number_style'),
                (stc.STC_LOUT_OPERATOR, 'operator_style'),
                (stc.STC_LOUT_STRING, 'string_style'),
                (stc.STC_LOUT_STRINGEOL, 'stringeol_style'),
                (stc.STC_LOUT_WORD, 'scalar_style'),
                (stc.STC_LOUT_WORD2, 'keyword2_style'),
                (stc.STC_LOUT_WORD3, 'keyword_style'),
                (stc.STC_LOUT_WORD4, 'class_style')]

#---- Extra Properties ----#
FOLD_COMPACT = ("fold.compact", '1')

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for LOUT""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_LOUT)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [LOUT_KW1, LOUT_KW2, LOUT_KW3]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD_COMPACT]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'#']



###############################################################################
# Name: EncodingDialogDemo.py                                                 #
# Purpose: Test and demo file for eclib.EncodingDialog                        #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2010 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Test file for testing the EncodingDialog.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: EncodingDialogDemo.py 63824 2010-04-02 01:10:44Z CJP $"
__revision__ = "$Revision: 63824 $"

#-----------------------------------------------------------------------------#
# Imports
import os
import sys
import wx

import IconFile

# Put local package on the path
#sys.path.insert(0, os.path.abspath('../../src'))
import eclib

#-----------------------------------------------------------------------------#

class TestPanel(wx.Panel):
    def __init__(self, parent, log):
        wx.Panel.__init__(self, parent, wx.ID_ANY, size=(500,500))

        # Attributes
        self.btnDlg = wx.Button(self, label="Show Encoding Dialog")
        self.log = log

        # Layout
        self.__DoLayout()

        # Event Handlers
        self.Bind(wx.EVT_BUTTON, self.OnShowDialogBtn)

    def __DoLayout(self):
        """Layout the panel"""
        vsizer = wx.BoxSizer(wx.VERTICAL)
        vsizer.Add(self.btnDlg, 0, wx.ALIGN_CENTER|wx.TOP, 20)
        self.SetSizer(vsizer)

    def OnShowDialogBtn(self, evt):
        """Show one of the test dialogs"""
        e_obj = evt.GetEventObject()
        if e_obj == self.btnDlg:
            dlg = eclib.EncodingDialog(self, msg="Select an encoding",
                                       title="Encoding Dialog",
                                       elist=eclib.GetAllEncodings(),
                                       default="utf-8")
            dlg.SetBitmap(IconFile.Home.GetBitmap())
            dlg.CenterOnParent()
            if dlg.ShowModal() == wx.ID_OK:
                enc = dlg.GetEncoding()
                self.log.write("Got Encoding: %s" % enc)
            dlg.Destroy()

#----------------------------------------------------------------------

def runTest(frame, nb, log):
    win = TestPanel(nb, log)
    return win

class TestLog:
    def __init__(self):
        pass

    def write(self, msg):
        print msg

#----------------------------------------------------------------------

overview = eclib.encdlg.__doc__
title = "EncodingDialog"

#-----------------------------------------------------------------------------#
if __name__ == '__main__':
    try:
        import run
    except ImportError:
        app = wx.PySimpleApp(False)
        frame = wx.Frame(None, title="EncodingDialog Demo")
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(TestPanel(frame, TestLog()), 1, wx.EXPAND)
        frame.CreateStatusBar()
        frame.SetSizer(sizer)
        frame.SetInitialSize((300, 300))
        frame.Show()
        app.MainLoop()
    else:
        run.main(['', os.path.basename(sys.argv[0])] + sys.argv[1:])



# -*- coding: utf-8 -*-
"""
    pygments.formatters
    ~~~~~~~~~~~~~~~~~~~

    Pygments formatters.

    :copyright: Copyright 2006-2010 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""
import os.path
import fnmatch

from pygments.formatters._mapping import FORMATTERS
from pygments.plugin import find_plugin_formatters
from pygments.util import ClassNotFound

ns = globals()
for fcls in FORMATTERS:
    ns[fcls.__name__] = fcls
del fcls

__all__ = ['get_formatter_by_name', 'get_formatter_for_filename',
           'get_all_formatters'] + [cls.__name__ for cls in FORMATTERS]


_formatter_alias_cache = {}
_formatter_filename_cache = []

def _init_formatter_cache():
    if _formatter_alias_cache:
        return
    for cls in get_all_formatters():
        for alias in cls.aliases:
            _formatter_alias_cache[alias] = cls
        for fn in cls.filenames:
            _formatter_filename_cache.append((fn, cls))


def find_formatter_class(name):
    _init_formatter_cache()
    cls = _formatter_alias_cache.get(name, None)
    return cls


def get_formatter_by_name(name, **options):
    _init_formatter_cache()
    cls = _formatter_alias_cache.get(name, None)
    if not cls:
        raise ClassNotFound("No formatter found for name %r" % name)
    return cls(**options)


def get_formatter_for_filename(fn, **options):
    _init_formatter_cache()
    fn = os.path.basename(fn)
    for pattern, cls in _formatter_filename_cache:
        if fnmatch.fnmatch(fn, pattern):
            return cls(**options)
    raise ClassNotFound("No formatter found for file name %r" % fn)


def get_all_formatters():
    """Return a generator for all formatters."""
    for formatter in FORMATTERS:
        yield formatter
    for _, formatter in find_plugin_formatters():
        yield formatter
